#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Parameter definition class and helper functions for evaluating parameters."""

import re
import copy
import numpy as np

from ufit import UFitError

__all__ = ['fixed', 'expr', 'overall', 'datapar', 'limited', 'delta',
           'Param', 'expr_namespace']


class fixed(str):
    """Mark the parameter value as fixed.

    The ``fixed()`` wrapper can also be omitted; any string means that a
    parameter as a fixed value that can include references to other parameters.
    """
    def __init__(self, p):
        str.__init__(self, p)

class expr(str):
    """Mark the parameter value as an expression.

    The ``expr()`` wrapper can also be omitted; any string means that a
    parameter as a fixed value that can include references to other parameters.
    """
    def __init__(self, e):
        str.__init__(self, e)

class overall(object):
    """Mark the parameter as an "overall" (global) parameter in a global fit.

    The argument can be another parameter definition, e.g. ``overall(limited(0,
    10, 2))``.
    """
    def __init__(self, v):
        self.v = v

class datapar(object):
    """Mark the parameter as coming from the data file's metadata.

    ``datapar('foo')`` is equivalent to ``expr('data.foo')``.
    """
    def __init__(self, v):
        self.v = v

class limited(tuple):
    """Give parameter limits together with the initial value.

    Example use::

       Gauss('peak', pos=0, ampl=limited(0, 100, 50), fwhm=1)
    """
    def __new__(self, min, max, v):
        return (min, max, v)

class delta(object):
    """Give parameter delta together with initial value."""
    def __init__(self, delta, v):
        self.delta = delta
        self.v = v


expr_namespace = {
    'data': None,  # replaced by the dataset's metadata dict, but in here
                   # so that no parameter can be called "data"
}
for fcn in ['pi', 'sqrt', 'sin', 'cos', 'tan', 'arcsin', 'arccos',
            'arctan', 'exp', 'log', 'radians', 'degrees', 'ceil',
            'floor', 'sinh', 'cosh', 'tanh']:
    expr_namespace[fcn] = getattr(np, fcn)

id_re = re.compile('[a-zA-Z_][a-zA-Z0-9_]*$')


class Param(object):
    def __init__(self, name, value=0, expr=None, pmin=None, pmax=None,
                 overall=False, delta=0, error=0, correl=None,
                 finalize=lambda x: x):
        if not id_re.match(name):
            raise UFitError('Parameter name %r is not a valid Python '
                            'identifier' % name)
        if name in expr_namespace:
            raise UFitError('Parameter name %r is reserved' % name)
        self.name = name
        self.value = value
        self.expr = expr
        self.pmin = pmin
        self.pmax = pmax
        # true if a global parameter for a global fit
        self.overall = overall
        # transform parameter after successful fit
        self.finalize = finalize
        # for backends that support setting parameter increments
        self.delta = delta
        # properties set on fit result
        self.error = error
        self.correl = correl or {}

    @classmethod
    def from_init(cls, name, pdef):
        if isinstance(pdef, cls):
            return pdef
        self = cls(name)
        while not isinstance(pdef, (int, long, float, str)):
            if isinstance(pdef, overall):
                self.overall = True
                pdef = pdef.v
            elif isinstance(pdef, datapar):
                self.expr = 'data.' + pdef.v
                pdef = 0
            elif isinstance(pdef, delta):
                self.delta = pdef.delta
                pdef = pdef.v
            elif isinstance(pdef, tuple) and len(pdef) == 3:
                self.pmin, self.pmax, pdef = pdef
            else:
                raise UFitError('Parameter definition %s not understood' %
                                pdef)
        if isinstance(pdef, str):
            self.expr = pdef
        else:
            self.value = pdef
        return self

    def copy(self, newname=None):
        cp = copy.copy(self)
        cp.name = newname or self.name
        cp.finalize = self.finalize
        return cp

    def fix(self):
        self.expr = str(self.value)

    def unfix(self):
        self.expr = None

    def __reduce__(self):
        return (Param, (self.name, self.value, self.expr, self.pmin,
                        self.pmax, self.overall, self.delta, self.error,
                        self.correl))

    def __str__(self):
        s = '%-15s = %10.5g +/- %10.5g' % (self.name, self.value, self.error)
        if self.expr:
            s += ' (fixed: %s)' % self.expr
        if self.overall:
            s += ' (global)'
        return s

    def __repr__(self):
        if self.expr:
            return '<Param %s = %.5g (expr: %s)>' % (
                self.name, self.value, self.expr)
        return '<Param %s = %.5g +/- %s>' % (
            self.name, self.value, self.error)

    def set_props(self, value, error, expr, pmin, pmax, delta):
        self.value = value
        self.error = error
        self.expr = expr
        self.pmin = pmin
        self.pmax = pmax
        self.delta = delta


# XXX replace by something more safe later
param_eval = eval

def prepare_params(params, meta):
    # find parameters that need to vary
    dependent = {}
    varying = []
    varynames = []
    for p in params:
        if p.expr:
            dependent[p.name] = [p.expr, None]
        else:
            varying.append(p)
            varynames.append(p.name)

    pd = dict((p.name, p.value) for p in varying)
    pd.update(expr_namespace)
    pd['data'] = meta

    # poor man's dependency tracking of parameter expressions
    dep_order = []
    maxit = len(dependent) + 1
    while dependent:
        maxit -= 1
        if maxit == 0:
            s = '\n'.join('   %s: %s' % (k, v[1]) for (k, v)
                          in dependent.iteritems())
            raise UFitError('Detected unresolved parameter dependencies:\n' + s)
        for p, (expr, _) in dependent.items():  # dictionary will change
            try:
                pd[p] = param_eval(expr, pd)
            except NameError, e:
                dependent[p][1] = str(e)
            except AttributeError, e:
                dependent[p][1] = 'depends on data.' + str(e)
            else:
                del dependent[p]
                dep_order.append((p, expr))
    #pd.pop('__builtins__', None)

    return varying, varynames, dep_order, pd


def update_params(parexprs, meta, pd):
    pd.update(expr_namespace)
    pd['data'] = meta
    for p, expr in parexprs:
        pd[p] = param_eval(expr, pd)
    #pd.pop('__builtins__', None)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Unified "from ufit.lab import *" module for quick scripts."""

from ufit import *
from pylab import *



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Embedded IPython qt console."""

from PyQt4.QtCore import SIGNAL
from PyQt4.QtGui import QMainWindow

from IPython.qt.console.rich_ipython_widget import RichIPythonWidget
from IPython.qt.inprocess import QtInProcessKernelManager


class QIPythonWidget(RichIPythonWidget):
    """Convenience class for a live IPython console widget."""
    def __init__(self, customBanner=None, *args, **kwargs):
        if customBanner is not None:
            self.banner = customBanner
        super(QIPythonWidget, self).__init__(*args, **kwargs)
        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel()
        kernel_manager.kernel.gui = 'qt4'
        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()
        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
            self.emit(SIGNAL('close'))
        self.exit_requested.connect(stop)
        def redraw():
            self.emit(SIGNAL('redraw'))
        self.executed.connect(redraw)

    def pushVariables(self, variableDict):
        """Given a dictionary containing name / value pairs, push those
        variables to the IPython console widget.
        """
        self.kernel_manager.kernel.shell.push(variableDict)

    def clearTerminal(self):
        """Clears the terminal."""
        self._control.clear()

    def printText(self, text):
        """Prints some plain text to the console."""
        self._append_plain_text(text)

    def executeCommand(self, command):
        """Execute a command in the frame of the console widget."""
        self._execute(command, True)


class ConsoleWindow(QMainWindow):
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        self.setWindowTitle('ufit Python console')
        self.ipython = QIPythonWidget(
            '''\
ufit interactive Python shell

Objects in the namespace:
* fig -- figure of the main viewport
* ax  -- current axes of the main viewport
* D   -- list of dataset items
            ''',
            self)
        self.setCentralWidget(self.ipython)

        self.connect(self.ipython, SIGNAL('close'), self.close)
        self.connect(self.ipython, SIGNAL('redraw'), parent.canvas.draw)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Models for several data corrections."""

from numpy import exp, sqrt, arcsin, tan, pi

from ufit.models import Model

__all__ = ['Background', 'SlopingBackground', 'CKI_Corr', 'Bose']


class Background(Model):
    """Constant background

    Parameters:

    * `bkgd` - the background (forced to be nonnegative)
    """
    param_names = ['bkgd']

    def __init__(self, name='', bkgd=0):
        pb, = self._init_params(name, self.param_names, locals())
        # background should be positive
        self.params[0].finalize = abs

        self.fcn = lambda p, x: abs(p[pb]) + 0*x

    def is_modifier(self):
        return True

    pick_points = ['background']

    def convert_pick(self, b):
        return {self.params[0].name: b[1]}


class SlopingBackground(Model):
    """Linearly sloping background

    Parameters:

    * `bkgd` - constant factor
    * `slope` - slope coefficient
    """
    param_names = ['bkgd', 'slope']

    def __init__(self, name='', bkgd=0, slope=0):
        pb, ps = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: p[pb] + x*p[ps]

    def is_modifier(self):
        return True

    pick_points = ['left background',
                   'right background']

    def convert_pick(self, b1, b2):
        slope = (b2[1] - b1[1]) / (b2[0] - b1[0])
        return {
            self.params[0].name: b1[1] - slope*b1[0],
            self.params[1].name: slope,
        }


class CKI_Corr(Model):
    """Correction for constant-k_i energy scans

    Parameters:

    * `ki` - the ki value in Ang-1
    * `dval` - the monochromator d-value in Ang
    """
    param_names = ['ki', 'dval']

    def __init__(self, name='', ki=None, dval='3.355'):
        pki, pdv = self._init_params(name, self.param_names, locals())
        def fcn(p, x):
            ki = p[pki]
            kf = sqrt(ki**2 - x/2.072)
            return kf**3/ki**3 * tan(arcsin(pi/ki/p[pdv]))/tan(arcsin(pi/kf/p[pdv]))
        self.fcn = fcn

    def is_modifier(self):
        return True


class Bose(Model):
    """Bose factor

    Parameters:

    * `tt` - the temperature in K
    """
    param_names = ['tt']

    def __init__(self, name='', tt=None):
        ptt, = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: x / (1. - exp(-11.6045*(x + 0.00001) / p[ptt]))

    def is_modifier(self):
        return True



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for TAIPAN data files."""

from os import path

from numpy import loadtxt, array


def guess_cols(colnames, coldata, meta):
    return meta['def_x'], meta['def_y'], None, meta['preset_channel']


def check_data(fp):
    line = fp.readline()
    fp.seek(0, 0)
    # on the first line is always name of the raw file
    return line.startswith('# raw_file =')


def read_data(filename, fp):
    line1 = ''
    line2 = fp.readline()
    skiprows = 0
    meta = {}
    meta['instrument'] = 'TAIPAN'
    # find the first non-comment line
    while line2.startswith('#'):
        # parse meta:
        line1 = line2
        line2 = fp.readline()
        skiprows += 1
        # do not parse headers
        if not line2.startswith('#'):
            continue
        key, oval = [x.strip() for x in line1[1:].strip().split('=', 1)]
        #name cannot start with col_ - this is reserved for columns values
        if key.startswith('col_'):
            key = key[4:]
        if key == 'experiment':
            title = oval
        elif key == 'experiment_number':
            meta['experiment'] = oval
        elif key == 'samplename':
            remark = oval
        if key == 'scan':
            meta['filenumber'] = int(oval)
        elif key == 'scan_title':
            meta['subtitle'] = oval
        else:
            meta[key] = oval
    if remark and title:
        meta['title'] = title + ', ' + remark
    print meta
    # now line2 is the first data line
    # line1 is header line
    line1 = line1[1:]
    # if there are comments, line1 will have the comment char
    colnames = line1.split()

    fp.seek(0, 0)
    arr = loadtxt(fp, ndmin=2, skiprows=skiprows, comments="#")
    # if number of colnames is not correct, discard them
    if len(colnames) != arr.shape[1]:
        colnames = ['Column %d' % i for i in range(1, arr.shape[1]+1)]

    cols = dict((name, arr[:,i]) for (i, name) in enumerate(colnames))
    meta['environment'] = []
    if 'en' in colnames:
        meta['hkle'] = arr[:,(3,4,5,1)]
        deviations = array([(cs.max()-cs.min()) for cs in meta['hkle'].T])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    for col in cols:
        meta[col] = cols[col].mean()
    for tcol in ['temp', 'TC1_sensorB', 'TC1_sensorA', 'TC1_sensorC', 'TC1_sensorD']:
        if tcol in cols:
            meta['environment'].append('T = %.3f K' % meta[tcol])
            break
    if 'MAG' in cols:
        if meta['MAG'] > 0:
            meta['environment'].append('B = %.3f K' % meta['B'])

    return colnames, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for ILL TAS data."""

from warnings import catch_warnings

from numpy import array, genfromtxt, atleast_2d

from ufit import UFitError


def check_data(fp):
    dtline = fp.readline()
    fp.seek(0, 0)
    return dtline.startswith('RRRRRRRRRRRR')


def guess_cols(colnames, coldata, meta):
    xg, yg, mg = None, None, None
    if 'CNTS' in colnames:
        yg = 'CNTS'
    if colnames[0] == 'QH':
        deviations = array([(cs.max()-cs.min()) for cs in coldata.T[:4]])
        xg = colnames[deviations.argmax()]
    else:
        xg = colnames[0]
    maxmon = 0
    for i, colname in enumerate(colnames):
        if colname.startswith('M') and colname[1:].isdigit():
            if coldata[:,i].sum() > maxmon:
                maxmon = coldata[:,i].sum()
                mg = colname
    return xg, yg, None, mg


def read_data(filename, fp):
    line = ''
    xcol = None
    meta = {}
    while line.strip() != 'DATA_:':
        if line.startswith('STEPS:'):
            parts = line[6:].strip().rstrip(',').split(', ')
            for part in parts:
                k, s = part.split('=')
                if float(s.strip()) != 0 or xcol is None:
                    xcol = k[1:]
        elif line.startswith('COMND:'):
            meta['subtitle'] = ' '.join(line[7:].rstrip().lower().split())
        elif line.startswith('TITLE:'):
            meta['title'] = line[6:].strip()
        elif line.startswith('FILE_:'):
            meta['filenumber'] = int(line[6:].strip())
        elif line.startswith('PARAM:'):
            parts = line[6:].strip().rstrip(',').split(',')
            for part in parts:
                k, s = part.split('=')
                meta[k.strip()] = float(s.strip())
        elif line.startswith('INSTR:'):
            meta['instrument'] = line[6:].strip().lower()
        elif line.startswith('EXPNO:'):
            meta['experiment'] = line[6:].strip().lower()
        line = fp.readline()
        if not line:
            break
    all_names = fp.readline().split()
    if not all_names:
        raise UFitError('No data columns found in in file %r' % filename)
    usecols = []
    names = []
    for i, name in enumerate(all_names):
        # XXX have to do flipper handling right
        if name in ('PNT', 'F1', 'F2'):
            continue
        names.append(name)
        usecols.append(i)
    # Berlin implementation adds "Finished ..." in the last line,
    # pretend that it is a comment
    with catch_warnings(True) as warnings:
        arr = atleast_2d(genfromtxt(fp, usecols=usecols, comments='F',
                                    invalid_raise=False))
    for warning in warnings:
        print '!!!', warning.message
    for i, n in enumerate(names):
        meta[n] = arr[:,i].mean()
    meta['environment'] = []
    if 'TT' in meta:
        meta['environment'].append('T = %.3f K' % meta['TT'])
    if 'MAG' in meta:
        meta['environment'].append('B = %.5f T' % meta['MAG'])
    if names[3] == 'EN':
        meta['hkle'] = arr[:,:4]
        deviations = array([(cs.max()-cs.min()) for cs in arr.T[:4]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    elif names[0] == 'QH':  # 2-axis mode
        meta['hkle'] = arr[:,:3]
        meta['hkle'] = array([(h, k, l, 0) for (h, k, l) in meta['hkle']])
        deviations = array([(cs.max()-cs.min()) for cs in arr.T[:4]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    if len(arr) == 0:
        raise UFitError('No data found in file %r' % filename)
    return names, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Backend using plain scipy leastsq."""

from __future__ import absolute_import
from numpy import sqrt, inf
from scipy.optimize import leastsq

from ufit.param import prepare_params, update_params
from ufit.utils import get_chisqr

__all__ = ['do_fit', 'backend_name']

backend_name = 'scipy'

def do_fit(data, fcn, params, add_kw):
    x, y, dy = data.fit_columns
    meta = data.meta
    varying, varynames, dependent, _ = prepare_params(params, meta)

    def leastsqfcn(params, data):
        pd = dict(zip(varynames, params))
        update_params(dependent, meta, pd)
        return (fcn(pd, x) - y) / dy

    initpars = []
    warned = False
    for p in varying:
        initpars.append(p.value)
        if (p.pmin is not None or p.pmax is not None) and not warned:
            print 'Sorry, scipy backend cannot handle parameter bounds.'
            warned = True

    try:
        res = leastsq(leastsqfcn, initpars, args=(data,), full_output=1, **add_kw)
    except Exception, e:
        return False, str(e), 0

    popt, pcov, infodict, errmsg, ier = res
    success = (ier in [1, 2, 3, 4])

    nfree = len(y) - len(varying)
    if nfree > 0 and pcov is not None:
        s_sq = (leastsqfcn(popt, data)**2).sum() / nfree
        pcov = pcov * s_sq
    else:
        pcov = inf

    pd = {}
    for i, p in enumerate(varying):
        pd[p.name] = popt[i]
        if pcov is not inf:
            p.error = sqrt(pcov[i,i])
        else:
            p.error = 0
        p.correl = {}  # XXX
    update_params(dependent, meta, pd)
    for p in params:
        p.value = pd[p.name]

    return success, errmsg, get_chisqr(fcn, x, y, dy, params)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for simple whitespace-separated column data files."""

from os import path

from numpy import loadtxt


def guess_cols(colnames, coldata, meta):
    if len(colnames) > 2:
        dycol = colnames[2]
    else:
        dycol = None
    return colnames[0], colnames[1], dycol, None


def check_data_simple(fp, sep=None):
    line = fp.readline()
    # find the first non-comment line
    while line.startswith(('#', '%')):
        line = fp.readline()
    line2 = fp.readline()
    fp.seek(0, 0)
    # must be values in non-comment line, or the line after
    try:
        map(float, line.split(sep))
    except ValueError:
        try:
            map(float, line2.split(sep))
        except ValueError:
            return False
    return True

def check_data(fp):
    return check_data_simple(fp, None)


def read_data_simple(filename, fp, sep=None):
    line1 = ''
    line2 = fp.readline()
    skiprows = 0
    # find the first non-comment line
    while line2.startswith(('#', '%')):
        line1 = line2
        line2 = fp.readline()
        skiprows += 1
    # now line2 is the first non-comment line (but may be column names)

    # if there are comments, line1 will have the comment char
    comments = '#'
    if line1.startswith(('#', '%')):
        comments = line1[0]
        line1 = line1[1:]
    try:
        map(float, line2.split())
    except ValueError:
        # must be column names
        colnames = line2.split()
        skiprows += 1
    else:
        # line1 might have column names
        if line1:
            colnames = line1.split()
        else:
            colnames = []
    fp.seek(0, 0)
    arr = loadtxt(fp, ndmin=2, skiprows=skiprows, comments=comments)
    # if number of colnames is not correct, discard them
    if len(colnames) != arr.shape[1]:
        colnames = ['Column %d' % i for i in range(1, arr.shape[1]+1)]
    meta = {}
    meta['filedesc'] = path.basename(filename)
    return colnames, arr, meta

def read_data(filename, fp):
    return read_data_simple(filename, fp, None)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for OLD NICOS data."""

import time

from ufit import UFitError
from ufit.data.nicos import _nicos_common_load

def check_data(fp):
    dtline = fp.readline()
    fp.seek(0, 0)
    # not sure if it is enough, but it is working
    return dtline.startswith('filename')

# guess_cols is the same as for new nicos format

from ufit.data.nicos import guess_cols


mapping = {
    'name': 'title',
    '1st orientation reflection': 'Sample_orient1',
    '2nd orientation reflection': 'Sample_orient2',
    'zone axis for scattering plane': 'Sample_zoneaxis',
    'created at': 'created',
    'PSI0 (deg)': 'Sample_psi0',
    'psi0 (deg)': 'Sample_psi0',
    'a,b,c (A)': 'Sample_lattice',
    'alpha,beta,gamma (deg)': 'Sample_angles',
    'Scattering sense': 'scatteringsense',
    'mono focussing mode': 'mono_focus',
    'ana  focussing mode': 'ana_focus',
    'TAS operation mode': 'opmode',
}

blacklist = set([
    'installation', 'sourcetype', 'sourcepower', 'moderator',
    'moderator temperature', 'beamtube', 'white flux@monochr.',
    'beam tube width', 'beam tube height', 'responsable',
    'phone', 'fax', 'mth  (A1) (deg)', 'mtt  (A2) (deg)',
    'psi  (A3) (deg)', 'phi  (A4) (deg)', 'ath  (A5) (deg)',
    'att  (A6) (deg)', 'sth  (A3) (deg)', 'stt  (A4) (deg)',
])

def read_data(filename, fp):
    meta = {}
    first_pos = fp.tell()
    dtline = fp.readline()
    fp.seek(first_pos)
    if not dtline.startswith('filename'):
        raise UFitError('%r does not appear to be an old NICOS data file' % filename)
    for line in iter(fp.readline, ''):
        # finished, go for data
        if line.startswith('scan data'):
            break
        # skip these lines
        if line.startswith(('***', '[', 'Sample information',
                            'instrument general setup at file creation',
                            'offsets of main axes')):
            continue
        #POLARIZATION, not implemented:
        #if line.startswith('counting for switching devices'): # polarized measurements!
        #    self.pol_devices = [d.strip() for d in line.split(']')[0].split('[')[1].split(', ')]
        #    self.pol_states = []
        #    self.polarized = True
        #    for s in line.split('states ')[1][1:-1].split('], ['):
        #        s = [d.strip() for d in s.split(', ')]
        #        ts = {}
        #        for i in range(len(self.pol_devices)):
        #            ts[self.pol_devices[i]] = s[i]
        #        self.pol_states.append(ts)

        try:
            key, value = line.split(':', 1)
        except ValueError:
            print 'ignored line:', repr(line)
            continue
        key = key.strip()
        value = value.strip()

        # some values are not important
        if key in blacklist:
            continue

        # some value names should be mapped
        if key in mapping:
            key = mapping[key]

        parts = value.split()
        if not parts:
            continue
        if key in ('ss1', 'ss2'):
            try:
                for i, side in enumerate(('left', 'right', 'bottom', 'top')):
                    meta['%s_%s' % (key, side)] = float(parts[i])
                meta[key] = tuple(float(b) for b in parts[:4])
            except Exception:
                continue
        elif value.endswith(('mm', 'deg', 'deg.', 'A-1', 'THz', 'meV', 'T', 'K',
                             'bar', '%', 's', 'min', 'min.', 'A')):
            try:
                meta[key] = float(parts[0])
            except ValueError:
                meta[key] = parts[0]
        else:
            meta[key] = value

    # convert some values
    if 'created' in meta:
        meta['created'] = time.mktime(time.strptime(meta['created'], '%m/%d/%Y %H:%M:%S'))
    if 'filename' in meta:
        meta['filename'] = meta['filename'].strip("'")
        meta['filenumber'] = int(meta['filename'].split("_")[1])

    # read data
    meta['subtitle'] = fp.readline().strip()
    colnames = fp.readline().split()
    colunits = fp.readline().split()

    return _nicos_common_load(fp, colnames, colunits, meta, '*')



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Fitting GUI windows for ufit."""

__all__ = ['start_fitter', 'start_loader']

from ufit.gui.loggers import getLogger
logger = getLogger('ufit')

from ufit.gui.fitter import start as start_fitter
from ufit.gui.dataloader import start as start_loader

import ufit.guiresource  # register icon resources



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data operations panel."""

from numpy import sqrt, ones

from PyQt4.QtCore import pyqtSignature as qtsig, SIGNAL
from PyQt4.QtGui import QWidget, QDialog, QMessageBox

from ufit.data.merge import rebin, floatmerge
from ufit.gui.common import loadUi
from ufit.gui.session import session


class DataOps(QWidget):

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.item = None
        self.data = None
        self.picking = None
        self.picked_points = []

        loadUi(self, 'dataops.ui')
        self.pickedLbl.hide()

    def initialize(self, item):
        self.data = item.data
        self.model = item.model
        self.item = item
        if self.data.fitmin is not None:
            self.limitminEdit.setText('%.5g' % self.data.fitmin)
        if self.data.fitmax is not None:
            self.limitmaxEdit.setText('%.5g' % self.data.fitmax)
        self.monscaleEdit.setText(str(self.data.nscale))
        self.titleEdit.setText(self.data.title)
        self.nameEdit.setText(self.data.name)

    def on_canvas_pick(self, event):
        if not hasattr(event, 'artist'):
            return
        if self.picking:
            xdata = event.artist.get_xdata()[event.ind]
            ydata = event.artist.get_ydata()[event.ind]
            self.picked_points.append(xdata)
            self.pickedLbl.setText('%d picked' % len(self.picked_points))
            event.canvas.figure.gca().plot([xdata], [ydata], 'ow', ms=8,
                                           mec='blue')
            event.canvas.draw()

    @qtsig('')
    def on_badResetBtn_clicked(self):
        self.data.reset_mask()
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_badPointsBtn_clicked(self):
        if self.picking == 'bad':
            self.badPointsBtn.setText('Start')
            self.pickedLbl.hide()
            self.picking = None
            self.removeBadPoints(self.picked_points)
        elif not self.picking:
            self.badPointsBtn.setText('Click points on plot, then '
                                      'here to finish')
            self.emit(SIGNAL('pickRequest'), self)
            self.picking = 'bad'
            self.picked_points = []
            self.pickedLbl.setText('0 picked')
            self.pickedLbl.show()

    @qtsig('')
    def on_limitsBtn_clicked(self):
        try:
            limitmin = float(self.limitminEdit.text())
        except ValueError:
            limitmin = None
        try:
            limitmax = float(self.limitmaxEdit.text())
        except ValueError:
            limitmax = None
        self.data.fitmin, self.data.fitmax = limitmin, limitmax
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    def removeBadPoints(self, points):
        """'Remove' bad data points (just mask them out)."""
        for point in points:
            self.data.mask[self.data.x == point] = False
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_rebinBtn_clicked(self):
        try:
            binsize = float(self.precisionEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Enter a valid precision.')
            return
        new_array = rebin(self.data._data, binsize)
        self.data.__init__(self.data.meta, new_array,
                           self.data.xcol, self.data.ycol, self.data.ncol,
                           self.data.nscale, name=self.data.name,
                           sources=self.data.sources)
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_floatmergeBtn_clicked(self):
        try:
            binsize = float(self.precisionEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Enter a valid precision.')
            return
        new_array = floatmerge(self.data._data, binsize)
        self.data.__init__(self.data.meta, new_array,
                           self.data.xcol, self.data.ycol, self.data.ncol,
                           self.data.nscale, name=self.data.name,
                           sources=self.data.sources)
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_cloneBtn_clicked(self):
        new_data = self.data.copy()
        new_model = self.model.copy()
        from ufit.gui.scanitem import ScanDataItem
        session.add_item(ScanDataItem(new_data, new_model), self.item.group)

    @qtsig('')
    def on_mulBtn_clicked(self):
        try:
            const = float(self.scaleConstEdit.text())
        except ValueError:
            return
        self.data.y *= const
        self.data.y_raw *= const
        self.data.dy *= const
        self.data.dy_raw *= const
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_addBtn_clicked(self):
        try:
            const = float(self.addConstEdit.text())
        except ValueError:
            return
        self.data.y += const
        self.data.y_raw += const * self.data.norm
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_scaleXBtn_clicked(self):
        try:
            const = float(self.scaleXConstEdit.text())
        except ValueError:
            return
        self.data.x *= const
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_shiftBtn_clicked(self):
        try:
            const = float(self.shiftConstEdit.text())
        except ValueError:
            return
        self.data.x += const
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_monscaleBtn_clicked(self):
        try:
            const = int(self.monscaleEdit.text())
        except ValueError:
            return
        self.data.nscale = const
        self.data.norm = self.data.norm_raw / const
        self.data.y = self.data.y_raw/self.data.norm
        self.data.dy = sqrt(self.data.y_raw)/self.data.norm
        self.data.yaxis = self.data.ycol + ' / %s %s' % (const, self.data.ncol)
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_titleBtn_clicked(self):
        self.data.meta.title = str(self.titleEdit.text())
        self.emit(SIGNAL('titleChanged'))
        session.set_dirty()
        self.emit(SIGNAL('replotRequest'), None)

    @qtsig('')
    def on_nameBtn_clicked(self):
        self.data.name = str(self.nameEdit.text())
        session.set_dirty()
        self.emit(SIGNAL('replotRequest'), None)

    @qtsig('')
    def on_subtractBtn_clicked(self):
        from ufit.gui.scanitem import ScanDataItem
        dlg = QDialog(self)
        loadUi(dlg, 'subtract.ui')
        data2obj = dlg.setList.populate(ScanDataItem)
        if dlg.exec_() != QDialog.Accepted:
            return
        witems = dlg.setList.selectedItems()
        if not witems:
            return
        try:
            prec = float(dlg.precisionEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Please enter a valid precision.')
            return
        bkgd_data = data2obj[witems[0].type()].data
        if not dlg.destructBox.isChecked():
            new_data = self.data.copy()
        else:
            new_data = self.data
        new_data.name = new_data.name + '-' + bkgd_data.name
        new_data.sources.extend(bkgd_data.sources)

        # Subtraction algorithm works as follows: for each point in the
        # background, the points in the original data with an X value within
        # the selected precision are looked up, and the Y value is subtracted.
        # An array of indices is kept so that from every original data point
        # background is subtracted at most once.

        # indices of data points not corrected
        ind_unused = ones(len(new_data.x), dtype=bool)
        for xb, yb, dyb, nb in bkgd_data._data:
            ind = ind_unused & (new_data.x >= xb - prec) & (new_data.x <= xb + prec)
            scale = new_data.norm_raw[ind]/nb
            new_data.y_raw[ind] -= scale * yb
            new_data.dy_raw[ind] = sqrt(new_data.dy_raw[ind]**2 + (scale * dyb)**2)
            ind_unused &= ~ind
        new_data.y = new_data.y_raw / new_data.norm
        new_data.dy = new_data.dy_raw / new_data.norm
        # mask out points from which no background has been subtracted
        new_data.mask &= ~ind_unused

        if not dlg.destructBox.isChecked():
            new_model = self.model.copy()
            from ufit.gui.scanitem import ScanDataItem
            session.add_item(ScanDataItem(new_data, new_model), self.item.group)
        else:
            self.emit(SIGNAL('replotRequest'), None)
            session.set_dirty()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Models package for ufit."""

from ufit.models.base import *
from ufit.models.peaks import *
from ufit.models.corr import *
from ufit.models.conv import *
from ufit.models.other import *
from ufit.models.sqwtas import *

from ufit.models import base, peaks, corr, other, sqwtas

__all__ = base.__all__ + peaks.__all__ + corr.__all__ + other.__all__ + \
    sqwtas.__all__

# Concrete models that can be used in the simplified GUI interface.

concrete_models = [
    GaussInt,
    Gauss,
    LorentzInt,
    Lorentz,
    Voigt,
    PseudoVoigt,
    DHO,
    Background,
    SlopingBackground,
    CKI_Corr,
    Bose,
    Const,
    StraightLine,
    Parabola,
    Cosine,
    Sinc,
    ExpDecay,
    PowerLaw,
    GaussianConvolution,
]



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Main window for the standalone GUI."""

from os import path

from PyQt4.QtCore import pyqtSignature as qtsig, Qt, SIGNAL, QModelIndex, \
    QByteArray
from PyQt4.QtGui import QMainWindow, QVBoxLayout, QMessageBox, QMenu, QIcon, \
    QFileDialog, QDialog, QAction, QActionGroup, QInputDialog, QSplitter, \
    QLabel, QCursor

from ufit import backends, __version__
from ufit.data.dataset import ScanData, ImageData
from ufit.gui import logger
from ufit.gui.common import MPLCanvas, MPLToolbar, SettingGroup, loadUi, \
    path_to_str
from ufit.gui.dialogs import ParamExportDialog
from ufit.gui.dataloader import DataLoader
from ufit.gui.itemlist import ItemListModel
from ufit.gui.inspector import InspectorWindow
from ufit.gui.annotations import AnnotationWindow
from ufit.gui.scanitem import ScanDataPanel, ScanDataItem
from ufit.gui.imageitem import ImageDataItem
from ufit.gui.session import session, temp_session, SessionItem, ItemGroup

max_recent_files = 6


class UFitMain(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self)

        self.itempanels = {}
        self.multipanels = {}
        self.current_panel = None
        self.inspector_window = None
        self.annotation_window = None

        self.sgroup = SettingGroup('main')
        loadUi(self, 'main.ui')

        self.connect(self.menuRecent, SIGNAL('aboutToShow()'),
                     self.on_menuRecent_aboutToShow)
        self.menuMoveToGroup = QMenu('Move selected to group', self)
        self.menuMoveToGroup.setIcon(QIcon(':/drawer-open.png'))
        self.connect(self.menuMoveToGroup, SIGNAL('aboutToShow()'),
                     self.on_menuMoveToGroup_aboutToShow)
        self.menuRemoveGroup = QMenu('Remove group', self)
        self.menuRemoveGroup.setIcon(QIcon(':/drawer--minus.png'))
        self.connect(self.menuRemoveGroup, SIGNAL('aboutToShow()'),
                     self.on_menuRemoveGroup_aboutToShow)
        self.menuRenameGroup = QMenu('Rename group', self)
        self.connect(self.menuRenameGroup, SIGNAL('aboutToShow()'),
                     self.on_menuRenameGroup_aboutToShow)

        # populate plot view
        layout2 = QVBoxLayout()
        layout2.setContentsMargins(0, 0, 0, 0)
        self.canvas = MPLCanvas(self, maincanvas=True)
        self.canvas.mpl_connect('button_press_event', self.on_canvas_pick)
        self.canvas.mpl_connect('pick_event', self.on_canvas_pick)
        self.canvas.mpl_connect('button_release_event', self.on_canvas_click)
        self.toolbar = MPLToolbar(self.canvas, self)
        self.toolbar.setObjectName('mainplottoolbar')
        self.connect(self.toolbar, SIGNAL('popoutRequested'),
                     self.on_actionPopOut_triggered)
        self.addToolBar(self.toolbar)
        layout2.addWidget(self.canvas)
        self.plotframe.setLayout(layout2)

        # session events
        self.connect(session, SIGNAL('itemsUpdated'),
                     self.on_session_itemsUpdated)
        self.connect(session, SIGNAL('itemAdded'), self.on_session_itemAdded)
        self.connect(session, SIGNAL('filenameChanged'),
                     self.on_session_filenameChanged)
        self.connect(session, SIGNAL('dirtyChanged'),
                     self.on_session_dirtyChanged)

        # create data loader
        self.dloader = DataLoader(self, self.canvas.plotter)
        self.connect(self.dloader, SIGNAL('newDatas'), self.on_dloader_newDatas)
        self.stacker.addWidget(self.dloader)
        self.current_panel = self.dloader

        # create item model
        self.itemlistmodel = ItemListModel()
        self.itemTree.setModel(self.itemlistmodel)
        self.itemlistmodel.reset()
        self.itemTree.addAction(self.actionMergeData)
        self.itemTree.addAction(self.actionRemoveData)
        self.itemTree.addAction(self.menuMoveToGroup.menuAction())
        self.connect(self.itemTree, SIGNAL('newSelection'),
                     self.on_itemTree_newSelection)

        # backend selector
        self.backend_group = QActionGroup(self)
        for backend in backends.available:
            action = QAction(backend.backend_name, self)
            action.setCheckable(True)
            if backends.backend.backend_name == backend.backend_name:
                action.setChecked(True)
            self.backend_group.addAction(action)
            self.menuBackend.addAction(action)
            self.connect(action, SIGNAL('triggered()'),
                         self.on_backend_action_triggered)

        # manage button
        menu = QMenu(self)
        menu.addAction(self.actionMergeData)
        menu.addAction(self.actionRemoveData)
        menu.addMenu(self.menuMoveToGroup)
        menu.addAction(self.actionReorder)
        menu.addSeparator()
        menu.addAction(self.actionNewGroup)
        menu.addMenu(self.menuRenameGroup)
        menu.addMenu(self.menuRemoveGroup)
        self.manageBtn.setMenu(menu)

        # right-mouse-button menu for canvas
        menu = self.canvasMenu = QMenu(self)
        menu.addAction(self.actionUnzoom)
        menu.addAction(self.actionHideFit)
        menu.addSeparator()
        menu.addAction(self.actionDrawSymbols)
        menu.addAction(self.actionConnectData)
        menu.addAction(self.actionSmoothImages)
        menu.addSeparator()
        menu.addAction(self.actionDrawGrid)
        menu.addAction(self.actionShowLegend)

        # restore window state
        with self.sgroup as settings:
            geometry = settings.value('geometry', QByteArray())
            self.restoreGeometry(geometry)
            windowstate = settings.value('windowstate', QByteArray())
            self.restoreState(windowstate)
            splitstate = settings.value('splitstate', QByteArray())
            self.splitter.restoreState(splitstate)
            vsplitstate = settings.value('vsplitstate', QByteArray())
            self.vsplitter.restoreState(vsplitstate)
            self.recent_files = settings.value('recentfiles', []) or []

    def on_dloader_newDatas(self, datas, ingroup):
        for group in session.groups:
            if group.name == ingroup:
                break
        else:
            group = session.add_group(ingroup)
        items = []
        for data in datas:
            if isinstance(data, ScanData):
                items.append(ScanDataItem(data))
            elif isinstance(data, ImageData):
                items.append(ImageDataItem(data))
            else:
                logger.warning('unknown data type: %s' % data)
        session.add_items(items, group)

    def on_session_itemsUpdated(self):
        # remove all panels whose item has vanished
        for item, panel in self.itempanels.items():
            if item not in session.all_items:
                self.stacker.removeWidget(panel)
                del self.itempanels[item]

    def on_session_itemAdded(self, item):
        # a single item has been added, show it
        self.itemTree.setCurrentIndex(self.itemlistmodel.index_for_item(item))

    def on_session_filenameChanged(self):
        if session.filename:
            self.setWindowTitle('ufit - %s[*]' % session.filename)
            self._add_recent_file(session.filename)
        else:
            self.setWindowTitle('ufit[*]')

    def on_session_dirtyChanged(self, dirty):
        self.setWindowModified(dirty)

    def _add_recent_file(self, fname):
        """Add to recent file list."""
        if not fname:
            return
        if fname in self.recent_files:
            self.recent_files.remove(fname)
        self.recent_files.insert(0, fname)
        if len(self.recent_files) > max_recent_files:
            self.recent_files.pop(-1)

    def on_menuRecent_aboutToShow(self):
        """Update recent file menu"""
        recent_files = []
        for fname in self.recent_files:
            if fname != session.filename and path.isfile(fname):
                recent_files.append(fname)
        self.menuRecent.clear()
        if recent_files:
            for i, fname in enumerate(recent_files):
                action = QAction('%d - %s' % (i+1, fname), self)
                self.connect(action, SIGNAL("triggered()"), self.load_session)
                action.setData(fname)
                self.menuRecent.addAction(action)
        self.actionClearRecent.setEnabled(len(recent_files) > 0)
        self.menuRecent.addSeparator()
        self.menuRecent.addAction(self.actionClearRecent)

    @qtsig('')
    def on_actionClearRecent_triggered(self):
        """Clear recent files list"""
        self.recent_files = []

    def select_new_panel(self, panel):
        if hasattr(self.current_panel, 'save_limits'):
            self.current_panel.save_limits()
        self.current_panel = panel
        self.stacker.setCurrentWidget(self.current_panel)

    def on_canvas_pick(self, event):
        if isinstance(self.current_panel, ScanDataPanel):
            self.current_panel.on_canvas_pick(event)

    def on_canvas_click(self, event):
        if event.button == 3 and not self.toolbar.mode: # right button
            self.canvasMenu.popup(QCursor.pos())

    @qtsig('')
    def on_loadBtn_clicked(self):
        self.select_new_panel(self.dloader)
        self.itemTree.setCurrentIndex(QModelIndex())

    @qtsig('')
    def on_actionNewGroup_triggered(self):
        name = QInputDialog.getText(self, 'ufit', 'Please enter a name '
                                    'for the new group:')[0]
        if not name:
            return
        session.add_group(name)

    def on_menuMoveToGroup_aboutToShow(self):
        self.menuMoveToGroup.clear()
        for group in session.groups:
            action = QAction(group.name, self)
            def move_to(group=group):
                items = self.selected_items()
                if not items:
                    return
                session.move_items(items, group)
                self.re_expand_tree()
            self.connect(action, SIGNAL('triggered()'), move_to)
            self.menuMoveToGroup.addAction(action)

    def on_menuRemoveGroup_aboutToShow(self):
        self.menuRemoveGroup.clear()
        for group in session.groups:
            action = QAction(group.name, self)
            def remove(group=group):
                session.remove_group(group)
                self.re_expand_tree()
            self.connect(action, SIGNAL('triggered()'), remove)
            self.menuRemoveGroup.addAction(action)

    def on_menuRenameGroup_aboutToShow(self):
        self.menuRenameGroup.clear()
        for group in session.groups:
            action = QAction(group.name, self)
            def rename(group=group):
                name = QInputDialog.getText(self, 'ufit', 'Please enter a new name '
                                            'for the group:', text=group.name)[0]
                if not name:
                    return
                session.rename_group(group, name)
            self.connect(action, SIGNAL('triggered()'), rename)
            self.menuRenameGroup.addAction(action)

    @qtsig('')
    def on_actionRemoveData_triggered(self):
        items = self.selected_items()
        if not items:
            return
        if QMessageBox.question(self, 'ufit',
                                'OK to remove %d item(s)?' % len(items),
                                QMessageBox.Yes|QMessageBox.No) == QMessageBox.No:
            return
        session.remove_items(items)
        self.re_expand_tree()

    @qtsig('')
    def on_actionReorder_triggered(self):
        dlg = QDialog(self)
        loadUi(dlg, 'reorder.ui')
        data2obj = dlg.itemList.populate()
        if not dlg.exec_():
            return
        new_structure = []
        for i in range(dlg.itemList.count()):
            new_index = dlg.itemList.item(i).type()
            obj = data2obj[new_index]
            if isinstance(obj, ItemGroup):
                new_structure.append((obj, []))
            else:
                if not new_structure:
                    QMessageBox.warning(self, 'ufit',
                                        'Reordering invalid: every data item '
                                        'must be below a group')
                    return
                new_structure[-1][1].append(obj)
        session.reorder_groups(new_structure)
        self.re_expand_tree()

    def selected_items(self, itemcls=SessionItem):
        """Return a list of selected items that belong to the given class."""
        items = (index.internalPointer()
                 for index in self.itemTree.selectedIndexes())
        if not itemcls:
            # This will also return ItemGroup objects!
            return list(items)
        return [item for item in items if isinstance(item, itemcls)]

    def on_itemTree_newSelection(self):
        items = self.selected_items(None)
        if len(items) == 1 and isinstance(items[0], ItemGroup):
            # a group is selected -- act as if we selected all items
            items = items[0].items
        if len(items) == 0:
            self.on_loadBtn_clicked()
        elif len(items) == 1:
            item = items[0]
            if item not in self.itempanels:
                panel = self.itempanels[item] = \
                        item.create_panel(self, self.canvas)
                self.stacker.addWidget(panel)
            else:
                panel = self.itempanels[item]
            self.select_new_panel(panel)
            panel.plot(panel.get_saved_limits())
            self.toolbar.update()
            if self.inspector_window and isinstance(item, ScanDataItem):
                self.inspector_window.setDataset(item.data)
        else:
            paneltype = type(items[0])
            if paneltype not in self.multipanels:
                panel = self.multipanels[paneltype] = \
                        items[0].create_multi_panel(self, self.canvas)
                self.stacker.addWidget(panel)
            else:
                panel = self.multipanels[paneltype]
            panel.initialize(items)
            self.select_new_panel(panel)
            panel.plot()

    def on_itemTree_expanded(self, index):
        index.internalPointer().expanded = True
        session.set_dirty()

    def on_itemTree_collapsed(self, index):
        index.internalPointer().expanded = False
        session.set_dirty()

    def re_expand_tree(self):
        # expand / collapse all groups according to last saved state
        for group in session.groups:
            if group.expanded:
                self.itemTree.expand(self.itemlistmodel.index_for_group(group))
            else:
                self.itemTree.collapse(self.itemlistmodel.index_for_group(group))

    @qtsig('')
    def on_actionInspector_triggered(self):
        if self.inspector_window:
            self.inspector_window.activateWindow()
            return
        self.inspector_window = InspectorWindow(self)
        def deref():
            self.inspector_window = None
        self.connect(self.inspector_window, SIGNAL('replotRequest'),
                     lambda: self.current_panel.plot(True))
        self.connect(self.inspector_window, SIGNAL('closed'), deref)
        if isinstance(self.current_panel, ScanDataPanel):
            self.inspector_window.setDataset(self.current_panel.item.data)
        self.inspector_window.show()

    @qtsig('')
    def on_actionAnnotations_triggered(self):
        if self.annotation_window:
            self.annotation_window.activateWindow()
            return
        self.annotation_window = AnnotationWindow(self)
        def deref():
            self.annotation_window = None
        self.connect(self.annotation_window, SIGNAL('closed'), deref)
        self.annotation_window.show()

    @qtsig('')
    def on_actionLoadData_triggered(self):
        self.on_loadBtn_clicked()

    def on_actionHideFit_toggled(self, on):
        self.canvas.plotter.no_fits = on
        self.current_panel.plot()

    def on_actionConnectData_toggled(self, on):
        self.canvas.plotter.lines = on
        self.current_panel.plot()

    def on_actionDrawSymbols_toggled(self, on):
        self.canvas.plotter.symbols = on
        self.current_panel.plot()

    def on_actionShowLegend_toggled(self, on):
        self.canvas.plotter.legend = on
        self.current_panel.plot()

    def on_actionDrawGrid_toggled(self, on):
        self.canvas.plotter.grid = on
        self.current_panel.plot()

    def on_actionSmoothImages_toggled(self, on):
        self.canvas.plotter.imgsmoothing = on
        self.current_panel.plot()

    def _get_export_filename(self, filter='ASCII text (*.txt)'):
        initialdir = session.props.get('lastexportdir', session.dirname)
        filename = QFileDialog.getSaveFileName(
            self, 'Select export file name', initialdir, filter)
        if filename == '':
            return ''
        expfilename = path_to_str(filename)
        session.props.lastexportdir = path.dirname(expfilename)
        return expfilename

    @qtsig('')
    def on_actionExportASCII_triggered(self):
        expfilename = self._get_export_filename()
        if expfilename:
            self.current_panel.export_ascii(expfilename)

    @qtsig('')
    def on_actionExportFIT_triggered(self):
        expfilename = self._get_export_filename()
        if expfilename:
            self.current_panel.export_fits(expfilename)

    @qtsig('')
    def on_actionExportPython_triggered(self):
        expfilename = self._get_export_filename('Python files (*.py)')
        if expfilename:
            self.current_panel.export_python(expfilename)

    @qtsig('')
    def on_actionExportParams_triggered(self):
        items = self.selected_items(ScanDataItem)
        dlg = ParamExportDialog(self, items)
        if dlg.exec_() != QDialog.Accepted:
            return
        expfilename = self._get_export_filename()
        if expfilename:
            try:
                dlg.do_export(expfilename)
            except Exception, e:
                logger.exception('While exporting parameters')
                QMessageBox.warning(self, 'Error', 'Could not export '
                                    'parameters: %s' % e)

    @qtsig('')
    def on_actionPrint_triggered(self):
        self.canvas.print_()

    @qtsig('')
    def on_actionPopOut_triggered(self):
        new_win = QMainWindow()
        canvas = MPLCanvas(new_win)
        toolbar = MPLToolbar(canvas, new_win)
        new_win.addToolBar(toolbar)
        splitter = QSplitter(Qt.Vertical, new_win)
        splitter.addWidget(canvas)
        try:
            from ufit.gui.console import QIPythonWidget
        except ImportError:
            lbl = QLabel('Please install IPython with qtconsole to '
                         'activate the console part of this window.', new_win)
            splitter.addWidget(lbl)
        else:
            iw = QIPythonWidget('interactive pylab plotting prompt', new_win)
            iw.pushVariables({'ax': canvas.axes})
            iw.executeCommand('from ufit.lab import *')
            iw.executeCommand('sca(ax)')
            iw.setFocus()
            self.connect(iw, SIGNAL('redraw'), canvas.draw_idle)
            splitter.addWidget(iw)
        new_win.setCentralWidget(splitter)
        self.current_panel.plot(limits=None, canvas=canvas)
        new_win.show()

    @qtsig('')
    def on_actionSavePlot_triggered(self):
        self.canvas.save_figure()

    @qtsig('')
    def on_actionUnzoom_triggered(self):
        self.toolbar.home()

    def check_save(self):
        if not self.isWindowModified():  # nothing there to be saved
            return True
        resp = QMessageBox.question(self, 'ufit',
            'Save current session?\n%s' % session.filename,
            QMessageBox.Yes|QMessageBox.No|QMessageBox.Cancel)
        if resp == QMessageBox.Yes:
            return self.save_session()
        elif resp == QMessageBox.No:
            return True
        return False

    @qtsig('')
    def on_actionNewSession_triggered(self):
        if not self.check_save():
            return
        session.clear()
        self.on_loadBtn_clicked()

    @qtsig('')
    def on_actionLoad_triggered(self):
        if not self.check_save():
            return
        initialdir = session.dirname
        if not initialdir:
            with self.sgroup as settings:
                initialdir = settings.value('loadfiledirectory', '')
        filename = QFileDialog.getOpenFileName(
            self, 'Select file name', initialdir, 'ufit files (*.ufit)')
        if filename == '':
            return
        self.load_session(path_to_str(filename))

    @qtsig('')
    def on_actionInsert_triggered(self):
        initialdir = session.dirname
        if not initialdir:
            with self.sgroup as settings:
                initialdir = settings.value('loadfiledirectory', '')
        filename = QFileDialog.getOpenFileName(
            self, 'Select file name', initialdir, 'ufit files (*.ufit)')
        if filename == '':
            return
        self.insert_session(path_to_str(filename))

    def load_session(self, filename=None):
        if not filename:
            # Recent files action
            action = self.sender()
            if isinstance(action, QAction):
                filename = action.data()
        try:
            session.load(filename)
            with self.sgroup as settings:
                settings.setValue('loadfiledirectory', path.dirname(filename))
        except Exception, err:
            logger.exception('Loading session %r failed' % filename)
            QMessageBox.warning(self, 'Error', 'Loading failed: %s' % err)
        else:
            self.re_expand_tree()
            # if there are annotations, show the window automatically
            if session.props.get('annotations'):
                self.on_actionAnnotations_triggered()

    def insert_session(self, filename):
        temp_session.load(filename)
        basename = path.basename(filename)
        if basename.endswith('.ufit'):
            basename = basename[:-5]
        # XXX slight HACK
        for group in temp_session.groups:
            session.add_group(group.name + ' (from %s)' % basename)
            for item in group.items:
                session.add_item(item)

    @qtsig('')
    def on_actionSave_triggered(self):
        self.save_session()

    @qtsig('')
    def on_actionSaveAs_triggered(self):
        self.save_session_as()

    def save_session(self):
        if session.filename is None:
            return self.save_session_as()
        try:
            session.save()
        except Exception, err:
            logger.exception('Saving session failed')
            QMessageBox.warning(self, 'Error', 'Saving failed: %s' % err)
            return False
        return True

    def save_session_as(self):
        initialdir = session.dirname
        filename = QFileDialog.getSaveFileName(
            self, 'Select file name', initialdir, 'ufit files (*.ufit)')
        if filename == '':
            return False
        session.set_filename(path_to_str(filename))
        try:
            session.save()
        except Exception, err:
            logger.exception('Saving session failed')
            QMessageBox.warning(self, 'Error', 'Saving failed: %s' % err)
            return False
        return True

    @qtsig('')
    def on_actionMergeData_triggered(self):
        items = self.selected_items(ScanDataItem)
        if len(items) < 2:
            return
        dlg = QDialog(self)
        loadUi(dlg, 'rebin.ui')
        if dlg.exec_():
            try:
                precision = float(dlg.precisionEdit.text())
            except ValueError:
                return
            datalist = [i.data for i in items]
            new_data = datalist[0].merge(precision, *datalist[1:])
            session.add_item(ScanDataItem(new_data), self.items[-1].group)

    @qtsig('')
    def on_actionQuit_triggered(self):
        self.close()

    def closeEvent(self, event):
        if not self.check_save():
            event.ignore()
            return
        event.accept()
        with self.sgroup as settings:
            settings.setValue('geometry', self.saveGeometry())
            settings.setValue('windowstate', self.saveState())
            settings.setValue('splitstate', self.splitter.saveState())
            settings.setValue('vsplitstate', self.vsplitter.saveState())
            settings.setValue('recentfiles', self.recent_files)

    @qtsig('')
    def on_actionAbout_triggered(self):
        dlg = QDialog(self)
        dlg.setWindowFlags(Qt.Dialog | Qt.FramelessWindowHint)
        loadUi(dlg, 'about.ui')
        dlg.lbl.setText(dlg.lbl.text().replace('VERSION', __version__))
        dlg.exec_()

    @qtsig('')
    def on_backend_action_triggered(self):
        backends.set_backend(str(self.sender().text()))



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for simple comma-separated column data files."""

from ufit.data.simple import guess_cols, read_data_simple, check_data_simple

def read_data(filename, fp):
    return read_data_simple(filename, fp, ',')

def check_data(fp):
    return check_data_simple(fp, ',')



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Routine for creating data mappings."""

import numpy as np
from numpy import array, mgrid, clip

from matplotlib.cbook import flatten


def bin_mapping(x, y, runs, usemask=True, log=False, xscale=1, yscale=1,
                interpolate=100, minmax=None):
    from scipy.interpolate import griddata as griddata_sp
    if usemask:
        xss = array(list(flatten(run['col_'+x][run.mask] for run in runs))) * xscale
        yss = array(list(flatten(run['col_'+y][run.mask] for run in runs))) * yscale
        if log:
            zss = list(flatten(np.log10(run.y)[run.mask] for run in runs))
        else:
            zss = list(flatten(run.y[run.mask] for run in runs))
    else:
        # XXX duplication
        xss = array(list(flatten(run['col_'+x] for run in runs))) * xscale
        yss = array(list(flatten(run['col_'+y] for run in runs))) * yscale
        if log:
            zss = list(flatten(np.log10(run.y) for run in runs))
        else:
            zss = list(flatten(run.y for run in runs))
    if minmax is not None:
        if log:
            minmax = map(np.log10, minmax)
        zss = clip(zss, minmax[0], minmax[1])
    interpolate = interpolate * 1j
    xi, yi = mgrid[min(xss):max(xss):interpolate,
                   min(yss):max(yss):interpolate]
    zi = griddata_sp(array((xss, yss)).T, zss, (xi, yi))
    return xss/xscale, yss/yscale, xi/xscale, yi/yscale, zi



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for NICOS2 data."""

import time
from numpy import array, loadtxt

from ufit import UFitError


def check_data(fp):
    dtline = fp.readline()
    fp.seek(0, 0)
    return dtline.startswith('### NICOS data file')


def _hkle_index(colnames):
    # find the index of the "h" column or -1 if it's not a qscan
    # NOTE: don't support "h" in a position other than 0 since this would
    # recognize all scans as Q scans even if the HKL device is just added
    # for monitoring purposes.
    #if 'h' in colnames:
    if colnames and colnames[0] == 'h':
        qhindex = 0  # colnames.index('h')
        if qhindex < len(colnames) - 3 and colnames[qhindex+3] == 'E':
            return qhindex
    return -1


def guess_cols(colnames, coldata, meta):
    xg, yg, mg = None, None, None
    qhindex = _hkle_index(colnames)
    if qhindex > -1:
        deviations = array([cs.max() - cs.min()
                            for cs in coldata.T[qhindex:qhindex+4]])
        xg = colnames[qhindex + deviations.argmax()]
    else:
        xg = colnames[0]
    maxmon = 0
    maxcts = 0
    for i, colname in enumerate(colnames):
        if colname.startswith('mon'):
            if coldata[:,i].sum() > maxmon:
                maxmon = coldata[:,i].sum()
                mg = colname
        if colname.startswith(('det', 'ctr', 'psd.roi')):
            if coldata[:,i].sum() > maxcts:
                maxcts = coldata[:,i].sum()
                yg = colname
    if yg is None:
        yg = colnames[1]
    return xg, yg, None, mg


def read_data(filename, fp):
    meta = {}
    dtline = fp.readline()
    if not dtline.startswith('### NICOS data file'):
        raise UFitError('%r does not appear to be a NICOS data file' %
                        filename)
    ctime = time.mktime(time.strptime(
        dtline[len('### NICOS data file, created at '):].strip(),
        '%Y-%m-%d %H:%M:%S'))
    meta['created'] = ctime
    remark = ''
    for line in iter(fp.readline, ''):
        line = line.decode('ascii', 'ignore').encode('ascii')
        if line.startswith('### Scan data'):
            break
        if line.startswith('# '):
            items = line.strip().split(None, 3)
            try:
                oval, unit = items[3].split(None, 1)
                val = float(oval)
            except (IndexError, ValueError):
                try:
                    oval = items[3]
                    val = float(oval)
                except ValueError:
                    val = items[3]
                except IndexError:
                    continue
                unit = None
            key = items[1]
            if key.endswith(('_offset', '_precision')):
                # we don't need these for fitting
                continue
            if key.endswith('_value'):
                key = key[:-6]
            if key.endswith('_instrument'):
                meta['instrument'] = oval.lower()
                continue
            elif key.endswith('_proposal'):
                meta['experiment'] = oval.lower()
            elif key.endswith('_samplename'):
                meta['title'] = oval
            elif key.endswith('_remark'):
                remark = oval
            elif key == 'number':
                meta['filenumber'] = int(oval)
                continue
            elif key == 'info':
                meta['subtitle'] = val
                continue
            meta[key] = val
    if remark and 'title' in meta:
        meta['title'] += ', ' + remark
    colnames = fp.readline()[1:].split()
    colunits = fp.readline()[1:].split()
    return _nicos_common_load(fp, colnames, colunits, meta, '#')


def _nicos_common_load(fp, colnames, colunits, meta, comments):
    def convert_value(s):
        try:
            return float(s)
        except ValueError:
            return 0.0
    cvdict = dict((i, convert_value) for i in range(len(colnames))
                  if colnames[i] != ';')
    colnames = [name for name in colnames if name != ';']
    colunits = [unit for unit in colunits if unit != ';']
    usecols = cvdict.keys()
    coldata = loadtxt(fp, converters=cvdict, usecols=usecols, ndmin=2,
                      comments=comments)
    if not coldata.size:
        raise UFitError('empty data file')
    cols = dict((name, coldata[:,i]) for (i, name) in enumerate(colnames))
    meta['environment'] = []
    for col in cols:
        meta[col] = cols[col].mean()
    for tcol in ['Ts', 'sT', 'T_ccr5_A', 'T_ccr5_B', 'sensor1']:
        if tcol in cols:
            meta['environment'].append('T = %.3f K' % meta[tcol])
            break
    if 'B' in cols:
        meta['environment'].append('B = %.3f K' % meta['B'])
    qhindex = _hkle_index(colnames)
    if qhindex > -1:
        meta['hkle'] = coldata[:,qhindex:qhindex+4]
        deviations = array([cs.max() - cs.min()
                            for cs in coldata.T[qhindex:qhindex+4]])
        xg = colnames[qhindex + deviations.argmax()]
        meta['hkle_vary'] = xg
    return colnames, coldata, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Backend using "unifit" leastsq algorithm."""

from __future__ import absolute_import

from copy import copy
from time import time
from numpy import sqrt, infty, ones, zeros, absolute, maximum, minimum, \
     finfo, corrcoef, nonzero, diag, isnan, dot
from numpy.linalg import svd, pinv

from ufit.param import prepare_params, update_params
from ufit.utils import get_chisqr

__all__ = ['do_fit', 'backend_name']

backend_name = 'unifit'

def do_fit(data, fcn, params, add_kw):
    x, y, dy = data.fit_columns
    meta = data.meta
    varying, varynames, dependent, _ = prepare_params(params, meta)

    def leastsqfcn(x, params):
        pd = dict(zip(varynames, params))
        update_params(dependent, meta, pd)
        return fcn(pd, x)

    initpars = []
    initdp = []
    warned = False
    for p in varying:
        initpars.append(p.value)
        initdp.append(p.delta or p.value/10. or 0.01)
        if (p.pmin is not None or p.pmax is not None) and not warned:
            print 'Sorry, unifit backend cannot handle parameter bounds.'
            warned = True
    try:
        res = __leastsq((x, y, dy), leastsqfcn, initpars, initdp,
                        **add_kw)
    except Exception, e:
        raise
        return False, str(e), 0

    success = res['converged']
    errmsg = res['errmsg']

    pd = {}
    for i, p in enumerate(varying):
        pd[p.name] = res['values'][i]
        p.error = res['errors'][i]
        p.correl = {}  # XXX
    update_params(dependent, meta, pd)
    for p in params:
        p.value = pd[p.name]

    return success, errmsg, get_chisqr(fcn, x, y, dy, params)


class FitError(Exception):
    pass

def __dfdp(x, f, p, dp, func):
    """
    Returns the partial derivatives of function 'func'.
    'x'(vect) is x axis values, 'y' is y values, 'p' and 'dp' are
    parameters and their variation.
    output 'df/dp' is a vector(or matrix) of partials varying of 'dp'.
    """

    y = zeros((len(x), len(p)))
    #pp = copy(p)
    dpp = copy(dp)

    for i in range(len(p)):
        pp = copy(p)
        if dpp[i] == 0.0:
            dpp[i] = 0.00001
        pp[i] = p[i] + dpp[i]
        t = func(x, pp)
        y[:,i] = (t-f)/dpp[i]
    return y

def __leastsq((x, y, wt), func, p, dp, niter=50, mode='print', eps=1e-3):
    """Levenberg-Marquardt nonlinear regression of func(x,p) to y(x).

             ---------------
    OPTIONS IN:
      niter (integ) : maximal number of iterations (50 is default).
      mode (string) : can be 'print' or 'silent'. In silent mode output is not written to standard out
                      but is saved in a message string.

    Author:  EF <manuf@ldv.univ-montp2.fr>, RM <rfm2@ds2.uh.cwru.edu>, AJ <jutan@charon.engga.uwo.ca>
    Description: Non Linear Least Square multivariable fit.

    Adapted for Octave : E.Farhi.   04/98   (manuf@ldv.univ-montp2.fr)
    Richard I. Shrager (301)-496-1122
    Modified by A.Jutan (519)-679-2111  jutan@charon.engga.uwo.ca
    Modified by Ray Muzic 14-Jul-1992   rfm2@ds2.uh.cwru.edu

    Version 3.beta    revised 07/97.
    Part of 'Spectral tools'. E.Farhi. 06/97

    Adapted for python: M. Janoschek 12/2008 (marc.janoschek@frm2.tum.de)
    Refrences:
    Bard, Nonlinear Parameter Estimation, Academic Press, 1974.
    Draper and Smith, Applied Regression Analysis, John Wiley and Sons, 1981.
    """
    #processing of input arguments
    options = zeros((len(p),2))            #desired fractional precision in parameter estimates.
    options[:,1] = ones(len(p)) * infty    #maximum fractional step change in parameter vector.
                        #not yet implemented standards are set (zero for precision, infty for max change)
    #instr_p = self.add_pars                                #additional parameters that are NOT fitted
    #func = self.__getfunc(self.func, self.add_pars)        #python function object of the form 'y=f(x,instr_p)'.

    stol = eps      #tolerance on square error sum fractional improvement (1e-3 is default)


    #cycle message (mainly for silent mode)
    cyclemsg = ''
    errmsg = ''

    if not all(dp):
        raise FitError('no dp given for some parameters')

    #check if data vector have the same length
    m = len(y)
    n = len(p)
    ##w = len(wt)
    sh = x.shape
    m1 = sh[0]
    ##if len(sh) > 1:
    ##    m2 = sh[1]
    ##else:
    ##    m2 = 1

    if m1 != m:
      raise FitError('Input(x)/output(y) data must have same number of rows!')

    if len(options) == 0:
        options = zeros((n,2))
        options[:,1] = ones(n) * infty
    else:
        nor, noc = options.shape

        if nor != n:
            raise FitError('Options and parameter matrices must have same number of rows!')

        if noc != 2:
            options = zeros((n,2))
            options[:,1] = ones(n) * infty

    pprec = options[:,0]
    maxstep = options[:,1]

    #apply parameter constraints
    ##p = self.__apply_constraints(p)

    #setup for iterations (with initial data)
    f = func(x, p)
    fbest = copy(f)              #initial function values
    pbest = copy(p)              #initial parameters
    r     = wt*(y-f)             #error (elementwise multiplication of columnvectors
    sbest = dot(r,r)
    ss = sbest
    nrm = zeros(n)
    chgbrev = infty*ones(n)
    kvg = False
    epsLlast = 1
    epstab = [0.1, 1, 1e2, 1e4, 1e6]

    #print header for fitting cycles if direct output was demanded by user
    if mode == 'print':
        print('Iteration  Time(s)  Residual Variance')
        print('=====================================')
        print('%5i    %7.2f    %8.3f' % (0, 0, sbest/len(y)))

    for iteration in range(niter):
        t0 = time()
        pprev = copy(pbest) #current paras
        #print pprev
        prt   = __dfdp(x, fbest, pprev, dp, func) #partials of func for x
        r     = wt*(y-fbest)        #error
        sprev = copy(sbest)        #square error
        sgoal = (1-stol)*sprev        #square error to achieve

        for j in range(n):        #compute partials norm
            if dp[j] == 0.0:
                nrm[j] = 0.0
            else:
                prt[:,j] = wt*prt[:,j]
                nrm[j]   = dot(prt[:,j],prt[:,j])
                if nrm[j] > 0:
                    nrm[j] = 1./sqrt(nrm[j])

            prt[:,j] = nrm[j]*prt[:,j]    #normalizes partials

        prt,s,v = svd(prt, full_matrices=False)        #prt=unit matr, s=eigenval : prt=prt*s*v.T
        v = v.T
        g = dot(prt.T,r)            #gradient by Gauss-Newton formula

        for k in range(len(epstab)):
            epsL = maximum(epsLlast*epstab[k],1e-7)
            se  = sqrt(s*s+epsL)
            gse = g/se
            chg = dot(v,gse)*nrm    #change on params
                        #check the change constraints and apply is necessary
            for l in range(n):
                if maxstep[l] == infty:
                    break
                chg[l] = maximum(chg[l], -1 * absolute(maxstep[l]*pprev[l]))
                chg[l] = minimum(chg[l], absolute(maxstep[l]*pprev[l]))

            aprec = absolute(pprec*pbest)
            if any(absolute(chg) > 0.1*aprec):
                #only worth evaluating function if there is some non-miniscule change
                p = chg+pprev

                #apply parameter constraints
                ##p = self.__apply_constraints(p)

                f = func(x, p)
                r = wt*(y-f)
                ss = dot(r,r)
                if ss < sbest:
                    pbest = copy(p)
                    fbest = copy(f)
                    sbest = copy(ss)

                if ss <= sgoal:
                    #print 'ss <= sgoal', ss, sgoal
                    break

        epsLlast = epsL
        if ss < finfo(float).eps:
            print 'Sum of squares within machine precision'
            break # machine precession

        aprec=absolute(pprec*pbest)
        if ((absolute(chg)).all() < aprec.all()) and ((absolute(chgprev)).all() < aprec.all()):
            kvg = True
            print 'Parameter changes converged to specified precision'
            break
        else:
            chgprev = chg
            if ss > sgoal:
                #print 'ss > sgoal in outer loop!'
                break

        dt = time()-t0
        cyclemsg = '%5i    %7.2f    %8.3f' % (iteration+1, dt, ss/len(y))
        if mode == 'print':
            print(cyclemsg)

    kvg = (sbest > sgoal)

    #set return values
    p = copy(pbest)
    f = copy(fbest)
    ss = copy(sbest)
    kvg = (kvg or (sbest <= finfo(float).eps) or kvg)

    if not kvg:
        errmsg += '** CONVERGENCE NOT ACHIEVED! **\n'
    else:
        print('===== Converged =====================')

    #Calculate R (Ref Draper & Smith p.46)
    r=corrcoef(y*wt,f*wt)
    r2=r*r.T
    ss = sum(wt*(f-y))/len(y)

    #calculate variance cov matrix and correlation matrix of parameters
    #re-evaluate the Jacobian at optimal values
    jac = __dfdp(x, fbest, pprev, dp, func)
    ##msk = nonzero(dp)[0]
    ##n = len(msk)        # dfdp(x,fbest,pprev,dp,func, instr_p)
    ##jac = jac[:,msk]        # use only fitted parameters

    # following section is Ray Muzic's estimate for covariance and correlation
    # assuming covariance of data is a diagonal matrix proportional to
    # diag(1/wt.^2).
    # cov matrix of data est. from Bard Eq. 7-5-13, and Row 1 Table 5.1

    Qinv = diag(wt*wt)
    Q = diag((0*wt+1)/(wt**2))

    resid = y-f            #unweighted residuals
    covr=dot(resid.T,dot(Qinv,dot(resid,Q)))/(m-n) # covariance of residuals
    Vy=1/(1-n/m)*covr          # Eq. 7-13-22, Bard    covariance of the data

    jtgjinv=pinv(dot(jac.T,dot(Qinv,jac))) #pinv = pseudo-inverse of matrix

    if not isnan(jtgjinv).all():
        # Eq. 7-5-13, Bard cov of parmater estimates
        covp = dot(jtgjinv,dot(jac.T,dot(Qinv,dot(Vy,dot(Qinv,dot(jac,jtgjinv))))))

        corp = ones((n,n))
        for k in range(n):
            for j in range(k,n):
                if covp[k,k]*covp[j,j]:
                    corp[k,j] = covp[k,j]/sqrt(abs(covp[k,k]*covp[j,j]))

    else:
        errmsg += "Warning : leasqr : couldn\'t compute squared Jacobian inverse matrix (no covp, no corp).\n"
        covp = zeros((n,n))
        corp = ones((n,n))

    deltap = sqrt(diag(covp))

    res = {}
    res['values'] = []
    res['errors'] = []

    for i in range(len(p)):
        res['values'].append(p[i])
        res['errors'].append(deltap[i])

    #work out chi2
    v = len(y) - len(nonzero(dp))
    ChiSq = sum((wt*(f-y))**2)/v

    #write back results:
    res['converged'] = kvg
    res['cycles'] = iteration

    res['covmat'] = covp
    res['corr'] = corp
    res['errmsg'] = errmsg
    res['rchi2'] = ChiSq
    res['rv'] = sbest/len(y)
    res['r2'] = r2
    res['ycalc'] = f

    return res



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data loading and treatment for ufit."""

from ufit import UFitError
from ufit.data import ill, nicos, nicos_old, simple, simple_csv, trisp, \
    llb, cascade, taipan
from ufit.data.loader import Loader
from ufit.data.dataset import Dataset, ScanData, ImageData, DatasetList

data_formats_scan = {
    'ill': ill,
    'nicos': nicos,
    'old nicos': nicos_old,
    'simple': simple,
    'simple comma-separated': simple_csv,
    'trisp': trisp,
    'llb': llb,
    'taipan': taipan,
}

data_formats_image = {
    'cascade': cascade,
}

data_formats = dict(kv for kv in (data_formats_scan.items() +
                                  data_formats_image.items()))

__all__ = ['Dataset', 'DatasetList', 'ScanData', 'ImageData', 'sets',
           'set_datatemplate', 'set_dataformat', 'read_data', 'as_data',
           'read_numors']


# simplified interface for usage in noninteractive scripts

global_loader = Loader()
sets = global_loader.sets

def set_datatemplate(template):
    """Set a new template for data file names.

    Normally, data file names consist of a fixed part and a sequential number.
    Therefore ufit constructs file names from a data template, which should
    contain a placeholder like ``%06d`` (for a 6-digit sequential number), and
    the actual file number given in the :func:`read_data` function.

    An example::

       set_datatemplate('/data/exp/data2012n%06d.dat')
       d1 = read_data(100)
       d2 = read_data(101)
       # etc.
    """
    global_loader.template = template

def set_dataformat(format):
    """Set the input data format.

    Normally ufit autodetects file formats, but this can be overridden using
    this function.  Data formats are:

    * ``'ill'`` - ILL TAS data format
    * ``'llb'`` - LLB binary TAS data format (known working for data from 1T and
      4F unpolarized)
    * ``'nicos'`` - NICOS data format
    * ``'old nicos'`` - NICOS 1.0 data format
    * ``'trisp'`` - FRM-II TRISP data format
    * ``'taipan'`` - ANSTO Taipan data format
    * ``'simple'`` - simple whitespace-separated multi-column files
    * ``'simple comma-separated'`` - simple comma-separated multi-column files
    """
    if format not in data_formats:
        raise UFitError('Unknown data format: %r, available formats are %s'
                        % (format, ', '.join(data_formats)))
    global_loader.format = format

def read_data(n, xcol='auto', ycol='auto', dycol=None, ncol=None, nscale=1):
    """Read a data file.  Returns a :class:`Dataset` object.

    :param xcol: X column name (or 1-based index)
    :param ycol: Y column name (or 1-based index)
    :param ycol: Y errors column name (or 1-based index); the default is to take
        the square root of the Y column as appropriate for counts
    :param ncol: normalization column name (or 1-based index); typically a beam
        monitor column
    :param nscale: scale for the normalization column; the Y data is determined
        as ``y[i] = y_raw[i] / ncol[i] * nscale``
    """
    return global_loader.load(n, xcol, ycol, dycol, ncol, nscale)

def as_data(x, y, dy, name=''):
    """Quickly construct a :class:`Dataset` object from three numpy arrays."""
    return Dataset.from_arrays(name or 'data', x, y, dy)

def read_numors(nstring, binsize, xcol='auto', ycol='auto',
                dycol=None, ncol=None, nscale=1):
    """Read a number of data files.  Returns a list of :class:`Dataset`\s.

    :param nstring: A string that gives file numbers, with the operators given
        below.
    :param binsize: Bin size when files need to be merged according to
        *nstring*.

    Other parameters as in :func:`read_data`.

    *nstring* can contain these operators:

    * ``,`` -- loads multiple files
    * ``-`` -- loads multiple sequential files
    * ``+`` -- merges multiple files
    * ``>`` -- merges multiple sequential files

    For example:

    * ``'10-15,23'`` loads files 10 through 15 and 23 in 7 separate datasets.
    * ``'10+11,23+24'`` loads two datasets consisting of files 10 and 11 merged
      into one set, as well as files 23 and 24.
    * ``'10>15+23'`` merges files 10 through 15 and 23 into one single dataset.
    * ``'10,11,12+13,14'`` loads four sets.
    """
    return global_loader.load_numors(nstring, binsize, xcol, ycol,
                                     dycol, ncol, nscale)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for TRISP spin-echo data."""

from numpy import array, loadtxt, zeros

from ufit import UFitError


def check_data(fp):
    dtline = fp.readline()
    fp.seek(0, 0)
    return dtline.startswith('pnt  ')


def guess_cols(colnames, coldata, meta):
    xg, yg, dyg, mg = None, 'c1', None, 'mon'
    if 'QH' in colnames:
        qhindex = colnames.index('QH')
        deviations = array([(cs.max()-cs.min())
                            for cs in coldata.T[qhindex:qhindex+4]])
        xg = colnames[qhindex + deviations.argmax()]
    elif colnames[0] == 'TC1' and colnames[1] == 'TC4':
        xg = 'TC4'
    else:
        xg = colnames[0]
    if 'c1_1' in colnames:
        mg = 'mon_1'
        yg = 'c1_1'
    return xg, yg, None, mg


def good_ycol(c):
    return c == 'c1'


def read_data(filename, fp):
    line = ''
    meta = {}
    infofp = open(filename[:-4] + '.log', 'rb')
    # first line in scan info
    line = infofp.readline()
    meta['subtitle'] = ' '.join(line.lower().split())
    meta['title'] = ''  # nothing here
    meta['instrument'] = 'trisp'
    while not line.startswith('Limits'):
        line = infofp.readline()
        if '-----' in line:
            continue
        parts = line.split()
        try:
            if len(parts) == 2:
                meta[parts[0]] = float(parts[1])
            elif len(parts) == 3:  # encoder target, value
                                   # so take value
                meta[parts[0]] = float(parts[2])
        except ValueError:
            pass
    names = fp.readline().split()
    pal = 'pal' in names
    # file with polarization analysis?
    if not names:
        raise UFitError('No data columns found in file %r' % filename)
    usecols = range(len(names))
    if names[0] == 'pnt':
        usecols = range(1, len(names))
        names = names[1:]
    arr = loadtxt(fp, ndmin=2, usecols=usecols)
    for i, n in enumerate(names):
        meta[n] = arr[:,i].mean()
    meta['environment'] = []
    if 'TTA' in meta:
        meta['environment'].append('T = %.3f K' % meta['TTA'])
    if len(arr) == 0:
        raise UFitError('No data found in file %r' % filename)
    if pal:
        if 'QH' not in names:
            raise UFitError('Polarization data without QHKLE not supported')
        nfixed = names.index('E')  # fixed columns (same for all PA points)
        pal_values = set(arr[:,0])
        npal = len(pal_values) # number of PA points
        names_new = names[1:nfixed+1]
        nvary = arr.shape[1] - nfixed - 1  # without pal and fixed columns
        arr_new = zeros((arr.shape[0]//npal, nfixed + nvary*npal))
        for pal_value in sorted(pal_values):
            for name in names[nfixed+1:]:
                names_new.append(name + '_%d' % pal_value)
            arr_new[:, nfixed+(pal_value-1)*nvary:nfixed+pal_value*nvary] = \
                arr[pal_value-1::npal, nfixed+1:]
        arr_new[:,:nfixed] = arr[::npal,1:nfixed+1]  # take fixed points from first PA point
        names = names_new
        arr = arr_new
    if names[0] == 'QH':
        meta['hkle'] = arr[:,:4]
        deviations = array([(cs.max()-cs.min()) for cs in arr.T[:4]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    return names, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

__version__ = '1.0.0'


def debug(str):
    if __debug__:
        print str

class UFitError(Exception):
    pass

from ufit.result import *
from ufit.param import *
from ufit.data import *
from ufit.models import *
from ufit.backends import *



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""A list view and model for different session items in the GUI."""

# parts borrowed from M. Janoschek' nfit2 GUI

from PyQt4.QtCore import Qt, QSize, SIGNAL, QAbstractItemModel, QModelIndex
from PyQt4.QtGui import QTreeView, QStyledItemDelegate, QTextDocument, QStyle, \
    QAbstractItemView, QListWidget, QListWidgetItem

from ufit.gui.session import session, ItemGroup
from ufit.gui.scanitem import ScanDataItem


class ItemListWidget(QListWidget):
    """Static view of all items, without model."""

    def populate(self, itemcls=None):
        data2obj = {}
        i = 0
        for group in session.groups:
            i += 1
            data2obj[i] = group
            wi = QListWidgetItem('Group: ' + group.name, self, i)
            if itemcls:
                wi.setFlags(Qt.NoItemFlags) # make it unselectable
            for item in group.items:
                i += 1
                data2obj[i] = item
                itemstr = '   %d - %s' % (item.index, item.title)
                if isinstance(item, ScanDataItem):
                    itemstr += ' (%s)' % item.data.meta.filedesc
                wi = QListWidgetItem(itemstr, self, i)
                if itemcls and not isinstance(item, itemcls):
                    wi.setFlags(Qt.NoItemFlags)
        return data2obj


class ItemTreeView(QTreeView):

    def __init__(self, parent):
        QTreeView.__init__(self, parent)
        self.header().hide()
        #self.setRootIsDecorated(False)
        #self.setStyleSheet("QTreeView::branch { display: none; }")
        self.setItemDelegate(ItemListDelegate(self))
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)

    def selectionChanged(self, selected, deselected):
        self.emit(SIGNAL('newSelection'))
        QTreeView.selectionChanged(self, selected, deselected)


class ItemListModel(QAbstractItemModel):

    def __init__(self):
        QAbstractItemModel.__init__(self)
        self.connect(session, SIGNAL('itemsUpdated'), self.reset)
        self.connect(session, SIGNAL('itemUpdated'), self.on_session_itemUpdated)
        self.connect(session, SIGNAL('itemAdded'), self.on_session_itemAdded)
        self.connect(session, SIGNAL('groupAdded'), self.on_session_groupAdded)
        self.connect(session, SIGNAL('groupUpdated'), self.on_session_groupUpdated)
        self.groups = session.groups

    def index_for_item(self, item):
        groupidx = self.groups.index(item.group)
        groupidx = self.index(groupidx, 0)
        return self.index(item.group.items.index(item), 0, groupidx)

    def index_for_group(self, group):
        return self.index(self.groups.index(group), 0)

    def on_session_itemUpdated(self, item):
        index = self.index_for_item(item)
        self.dataChanged.emit(index, index)

    def on_session_groupUpdated(self, group):
        index = self.index(self.groups.index(group), 0)
        self.dataChanged.emit(index, index)

    def on_session_itemAdded(self, item):
        groupindex = self.index(self.groups.index(item.group), 0)
        self.dataChanged.emit(groupindex, groupindex)
        itemrow = item.group.items.index(item)
        self.rowsInserted.emit(groupindex, itemrow, itemrow)

    def on_session_groupAdded(self, group):
        itemrow = self.groups.index(group)
        self.rowsInserted.emit(QModelIndex(), itemrow, itemrow)

    def columnCount(self, parent=QModelIndex()):
        return 1

    def rowCount(self, index=QModelIndex()):
        if index.isValid():   # data items
            obj = index.internalPointer()
            if isinstance(obj, ItemGroup):
                return len(self.groups[index.row()].items)
            return 0
        return len(self.groups)

    def index(self, row, column, parent=QModelIndex()):
        if parent.isValid():  # data items
            group = parent.internalPointer()
            return self.createIndex(row, column, group.items[row])
        return self.createIndex(row, column, self.groups[row])

    def parent(self, index):
        obj = index.internalPointer()
        if isinstance(obj, ItemGroup):
            return QModelIndex()
        group = obj.group
        return self.createIndex(self.groups.index(group), 0, group)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        obj = index.internalPointer()
        if role == Qt.DisplayRole:
            return obj.htmldesc
        elif role == Qt.TextAlignmentRole:
            return int(Qt.AlignLeft|Qt.AlignVCenter)
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.TextAlignmentRole:
            if orientation == Qt.Horizontal:
                return int(Qt.AlignLeft|Qt.AlignVCenter)
            return int(Qt.AlignRight|Qt.AlignVCenter)
        if role != Qt.DisplayRole:
            return None


class ItemListDelegate(QStyledItemDelegate):

    def paint(self, painter, option, index):
        text = index.model().data(index)
        palette = self.parent().palette()
        document = QTextDocument()
        document.setDefaultFont(option.font)
        if option.state & QStyle.State_Selected:
            document.setHtml("<font color=%s>%s</font>" %
                    (palette.highlightedText().color().name(), text.decode('utf-8', 'ignore')))
            color = palette.highlight().color()
        else:
            document.setHtml(text)
            color = palette.base().color()
        painter.save()
        painter.fillRect(option.rect, color)
        painter.translate(option.rect.x(), option.rect.y())
        document.drawContents(painter)
        painter.restore()

    def sizeHint(self, option, index):
        text = index.model().data(index)
        document = QTextDocument()
        document.setDefaultFont(option.font)
        document.setHtml(text)
        return QSize(document.idealWidth(), int(document.size().height()))



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data merging routines."""

from numpy import arange, ones, zeros, sqrt, array

from ufit import UFitError


def rebin(data, binsize):
    """Simple rebinning of (x, y, dy, n) data."""
    # XXX make it work with hkle 4-d x data

    if binsize == 0:
        # no merging, just concatenate
        return data

    x, y, dy, n = data.T

    # calculate new x values

    halfbinsize = binsize/2.
    stops = arange(x.min() - (x.min() % binsize) - binsize,
                   x.max() - (x.max() % binsize) + 2*binsize,
                   binsize) + halfbinsize
    nbins = len(stops)

    # newarray will be the new x, y, dy, n array
    newarray = zeros((nbins, 4))
    newarray[:,0] = stops

    # this will keep track which data values we already used
    data_unused = ones(len(x), bool)

    # this will keep track which new bins are used; unused ones are
    # left out
    new_used = ones(nbins, bool)

    for i in xrange(nbins):
        stop = newarray[i, 0]
        # get indices of all data points with x values lying below stop
        indices = x <= (stop + halfbinsize)
        # remove indices of data already used in previous bins
        indices &= data_unused
        if indices.any():
            newarray[i, 1] += y[indices].sum()
            newarray[i, 2] += sqrt((dy[indices]**2).sum())
            newarray[i, 3] += n[indices].sum()
            data_unused[indices] = False
        else:
            new_used[i] = False

    # are there any data points left unused?
    if data_unused.any():
        raise UFitError('Merging data failed (data left over), check merging '
                        'algorithm for bugs')

    # remove any stops without monitor data
    newarray = newarray[new_used]

    # return array
    return newarray

def mergeList(tomerge):
    tomerge = array(tomerge)
    merged = tomerge.sum(axis = 0)
    merged[2] = sqrt((tomerge**2).sum(axis = 0)[2])
    merged[0] = merged[0] / len(tomerge)
    return merged;

def floatmerge(data, binsize):
    """Merging data based on floating window."""

    if binsize == 0:
        # no merging, just concatenate
        return data

    # sort data
    data = data[data[:,0].argsort()]

    lastvals = None
    tomerge =  []
    newlist =  []
    for vals in data:
        if lastvals != None:
            if vals[0] > lastvals[0] + binsize:
                #merge points in list:
                newlist.append(mergeList(tomerge))
                tomerge = []
        tomerge.append(vals)
        lastvals = vals
    newlist.append(mergeList(tomerge))

    # return array
    return array(newlist)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Various dialogs."""

from numpy import array, savetxt, nan

from PyQt4.QtCore import pyqtSignature as qtsig
from PyQt4.QtGui import QDialog, QListWidgetItem

from ufit.gui.common import loadUi
from ufit.data.dataset import Dataset


class ParamExportDialog(QDialog):
    def __init__(self, parent, items):
        QDialog.__init__(self, parent)
        loadUi(self, 'paramselect.ui')
        self.availList.populate(items, intersect=False)
        self.items = items

    def on_availList_itemDoubleClicked(self, item):
        QListWidgetItem(item.text(), self.selectList, item.type())

    @qtsig('')
    def on_addBtn_clicked(self):
        item = self.availList.currentItem()
        if not item:
            return
        QListWidgetItem(item.text(), self.selectList, item.type())

    @qtsig('')
    def on_removeBtn_clicked(self):
        item = self.selectList.currentItem()
        if not item:
            return
        self.selectList.takeItem(self.selectList.row(item))

    def do_export(self, filename):
        data = []
        colnames = [self.selectList.item(i).text().strip()
                    for i in range(self.selectList.count())]
        coltypes = [self.selectList.item(i).type()
                    for i in range(self.selectList.count())]
        errors = self.errorBox.isChecked()
        for item in self.items:
            drow = []
            for coltype, name in zip(coltypes, colnames):
                if coltype == 1:   # parameter
                    if name in item.model.paramdict:
                        drow.append(item.model[name].value)
                        if errors:
                            drow.append(item.model[name].error)
                    else:
                        drow.append(nan)
                        if errors:
                            drow.append(nan)
                else:    # data value
                    if name in item.data.meta:
                        drow.append(item.data.meta[name])
                    else:
                        drow.append(nan)
            data.append(drow)
        data = array(data)
        if self.headerBox.isChecked():
            header = []
            for coltype, name in zip(coltypes, colnames):
                header.append(name)
                if coltype == 1 and self.errorBox.isChecked():
                    header.append(name + '_error')
            header = '\t'.join(header)
        else:
            header = ''
        savetxt(filename, data, header=header)


class ParamSetDialog(QDialog):
    def __init__(self, parent, items):
        QDialog.__init__(self, parent)
        loadUi(self, 'paramset.ui')
        self.new_data = None
        self.items = items
        self.xaxisList.populate(items)
        self.yaxisList.populate(items)
        self._auto_name = ''

    def _gen_name(self):
        xi = self.xaxisList.currentItem()
        yi = self.yaxisList.currentItem()
        if self.nameBox.text() != self._auto_name:
            return
        auto_name = ''
        if yi:
            auto_name += yi.text().strip()
        if xi:
            auto_name += ' vs. ' + xi.text().strip()
        self._auto_name = auto_name
        self.nameBox.setText(auto_name)

    def on_xaxisList_itemSelectionChanged(self):
        self._gen_name()

    def on_yaxisList_itemSelectionChanged(self):
        self._gen_name()

    def exec_(self):
        res = QDialog.exec_(self)
        if res != QDialog.Accepted:
            return res
        xx, yy, dy = [], [], []
        xp = yp = False
        xv = self.xaxisList.currentItem().text().strip()
        if self.xaxisList.currentItem().type() == 1:
            xp = True
        yv = self.yaxisList.currentItem().text().strip()
        if self.yaxisList.currentItem().type() == 1:
            yp = True

        for item in self.items:
            if xp:
                xx.append(item.model.paramdict[xv].value)
            else:
                xx.append(item.data.meta[xv])
            if yp:
                yy.append(item.model.paramdict[yv].value)
                dy.append(item.model.paramdict[yv].error)
            else:
                yy.append(item.data.meta[yv])
                dy.append(1)
        xx, yy, dy = map(array, [xx, yy, dy])

        self.new_data = Dataset.from_arrays(str(self.nameBox.text()),
                                            xx, yy, dy, xcol=xv, ycol=yv)
        return res



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""ufit base models."""

import re
import inspect
import operator
import cPickle as pickle
from numpy import concatenate

from ufit import param, backends, UFitError, Param, Dataset
from ufit.result import Result, MultiResult
from ufit.utils import get_chisqr, cached_property
from ufit.plotting import DataPlotter

__all__ = ['Model', 'CombinedModel', 'Function', 'Custom', 'eval_model']


data_re = re.compile(r'\bdata\b')

def eval_model(modeldef, paramdef=None):
    from ufit import models
    d = models.__dict__.copy()
    d.update(param.expr_namespace)
    d.update(param.__dict__)
    model = eval(modeldef, d)
    model.python_code = modeldef
    if paramdef:
        model.params = paramdef
    return model


class Model(object):
    """Base class for Model functions.

    Important APIs:

    * fit() - fit data with the model
    * add_params() - add parameters that are referenced in parameter
      expressions but not given by a parameter of one of the models yet
    * get_components() - return a list of Model instances that represent
      individual components of the complete model
    * is_modifier() - return bool whether the specific model is a "modifier"
      (i.e. not a component)
    """

    # class properties

    # if nonempty, the names of points to pick in the GUI
    pick_points = []
    # names of parameters
    param_names = []

    # set by initializers
    name = ''
    params = []
    fcn = None
    _orig_params = None

    # can be set if the model is generated by eval()
    python_code = None

    # number of samples for plotting
    nsamples = 1000

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.name)

    def _init_params(self, name, pnames, init):
        """Helper for model subclasses to quickly initialize Param objects.

        If a model name is given by the user, the parameter names are prefixed
        with "name_", so that multiple parameters with the same name can
        coexist in the same model.
        """
        self.params = []
        self.name = name
        if name:
            pnames_real = ['%s_%s' % (name, pname) for pname in pnames]
        else:
            pnames_real = pnames
        for (pname, porigname) in zip(pnames_real, pnames):
            try:
                initval = init[porigname]
                if initval is None:
                    raise KeyError
                self.params.append(Param.from_init(pname, initval))
            except KeyError:
                # not raising an exception allows the GUI to omit irrelevant
                # initializers
                self.params.append(Param.from_init(pname, 0))
                #raise UFitError('Parameter %s needs an initializer' % pname)
        return pnames_real

    def _combine_params(self, a, b):
        """Helper for model subclasses that combine two submodels.

        self.params is initialized with a combination of params of both models,
        while an error is raised if name clash.
        """
        self.params = []
        seen = set()
        for m in [a, b]:
            for p in m.params:
                if p.name in seen:
                    raise UFitError('Parameter name clash: %s - give all model '
                                    'classes a name to avoid this' % p.name)
                seen.add(p.name)
                self.params.append(p)

    @cached_property
    def paramdict(self):
        return dict((p.name, p) for p in self.params)

    def __getitem__(self, key):
        return self.paramdict[key]

    def __add__(self, other):
        if isinstance(other, (int, long, float)):
            other = Constant(other)
        elif not isinstance(other, Model):
            return NotImplemented
        return CombinedModel(self, other, '+')

    def __radd__(self, other):
        if isinstance(other, (int, long, float)):
            return CombinedModel(Constant(other), self, '+')
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, (int, long, float)):
            other = Constant(other)
        elif not isinstance(other, Model):
            return NotImplemented
        return CombinedModel(self, other, '-')

    def __rsub__(self, other):
        if isinstance(other, (int, long, float)):
            return CombinedModel(Constant(other), self, '-')
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, long, float)):
            other = Constant(other)
        elif not isinstance(other, Model):
            return NotImplemented
        return CombinedModel(self, other, '*')

    def __rmul__(self, other):
        if isinstance(other, (int, long, float)):
            return CombinedModel(Constant(other), self, '*')
        return NotImplemented

    def __div__(self, other):
        if isinstance(other, (int, long, float)):
            other = Constant(other)
        elif not isinstance(other, Model):
            return NotImplemented
        return CombinedModel(self, other, '/')

    def __rdiv__(self, other):
        if isinstance(other, (int, long, float)):
            return CombinedModel(Constant(other), self, '/')
        return NotImplemented

    def __pow__(self, other):
        if isinstance(other, (int, long, float)):
            other = Constant(other)
        elif not isinstance(other, Model):
            return NotImplemented
        return CombinedModel(self, other, '**')

    @property
    def original_params(self):
        if self._orig_params is None:
            return self.params
        return self._orig_params

    def fit(self, data, **kw):
        """Fit the model to the data.  *data* must be a :class:`Dataset` object.

        Any keywords will be passed to the raw fitting routine of the backend.
        """
        if self._orig_params is None:
            self._orig_params = [p.copy() for p in self.params]
        # keeping the attribute chain like this allows the backend to
        # be changed on the fly
        success, msg, chi2 = backends.backend.do_fit(data, self.fcn,
                                                     self.params, kw)
        for p in self.params:
            p.value = p.finalize(p.value)
        return Result(success, data, self, self.params, msg, chi2)

    def global_fit(self, datas, **kw):
        """Fit the model to multiple datasets, given as a list by *datas*.

        Any keywords will be passed to the raw fitting routine of the backend.
        """
        return GlobalModel(self, datas).fit(datas, **kw)

    def multi_fit(self, datas, **kw):
        results = []
        for data in datas:
            self.reset()
            results.append(self.fit(data, **kw))
        return MultiResult(results)

    def reset(self):
        if self._orig_params is not None:
            self.params = [p.copy() for p in self._orig_params]

    def plot(self, data, axes=None, labels=True, pdict=None, **kw):
        """Plot the model and the data in the current figure."""
        DataPlotter(axes=axes).plot_model(self, data, labels, pdict, **kw)

    def plot_components(self, data, axes=None, labels=True, pdict=None, **kw):
        """Plot subcomponents of the model in the current figure."""
        DataPlotter(axes=axes).plot_model_components(self, data, labels,
                                                     pdict, **kw)

    def add_params(self, **params):
        """Add parameters that referenced by expressions in other parameters.

        For example, in this model ::

           m = Gauss('p1', pos='delta', ampl=5, fwhm=0.5) + \\
               Gauss('p2', pos='-delta', ampl='p1_ampl', fwhm='p1_fwhm')

        the parameter "delta" is referenced by two parameter expressions, but
        does not appear as a parameter of any of the model functions.  This
        parameter must be made known to the model by calling e.g. ::

           m.add_params(delta=0)
        """
        for pname, initval in params.iteritems():
            self.params.append(Param.from_init(pname, initval))

    def get_components(self):
        """Return a list of invidual non-modifier components.

        Modifiers are applied to the components as appropriate.
        """
        return [self]

    def is_modifier(self):
        """Return true if the model is a "modifier", i.e. not a component that
        should be plotted as a separate component.
        """
        return False

    def get_description(self):
        """Get a Python description of the model (no parameters)."""
        if self.python_code:
            return self.python_code
        if self.name:
            return '%s(%r)' % (self.__class__.__name__, self.name)
        return '%s()' % self.__class__.__name__

    def __reduce__(self):
        """Pickling support: reconstruct the object from a constructor call."""
        if self.python_code:
            return (eval_model, (self.python_code, self.params))
        return (self.__class__, (self.name,) + tuple(self.params))

    def copy(self):
        if self.python_code:
            return eval_model(self.python_code, [p.copy() for p in self.params])
        return pickle.loads(pickle.dumps(self))

    def get_pick_points(self):
        """Get a list of point names that should be picked for initial guess."""
        if self.name:
            return ['%s: %s' % (self.name, pn) for pn in self.pick_points]
        return self.pick_points

    def convert_pick(self, *args):
        """Convert pick point coordinates (x,y) to parameter initial guesses."""
        return {}

    def apply_pick(self, points):
        initial_values = self.convert_pick(*points)
        for p in self.params:
            if p.name in initial_values:
                p.value = initial_values[p.name]

    def export_python(self, fp, objname='model'):
        fp.write('%s = %s\n' % (objname, self.get_description()))
        for param in self.params:
            fp.write('%s[%r].set_props(%r, %r, %r, %r, %r, %r)\n' %
                     (objname, param.name, param.value, param.error,
                      param.expr, param.pmin, param.pmax, param.delta))


class CombinedModel(Model):
    """Models an arithmetic combination of two sub-models.

    Parameters are combined from both; their names may not clash.
    """

    op_prio = {
        '+': 0,
        '-': 0,
        '*': 1,
        '/': 1,
        '**': 2,
    }

    op_fcn = {
        '+':  operator.add,
        '-':  operator.sub,
        '*':  operator.mul,
        '/':  operator.div,
        '**': operator.pow,
    }

    def __init__(self, a, b, opstr):
        self.params = []
        self._a = a
        self._b = b
        self._op = op = self.op_fcn[opstr]
        self._opstr = opstr
        if a.name and b.name:
            self.name = a.name + opstr + b.name
        else:
            self.name = a.name or b.name
        self._combine_params(a, b)

        self.fcn = lambda p, x: op(a.fcn(p, x), b.fcn(p, x))

        # cache this!
        self._components = None

    def __repr__(self):
        return '<%s %r %s %r>' % (self.__class__.__name__,
                                  self._a, self._opstr, self._b)

    def __reduce__(self):
        """Pickling support: reconstruct the object from a constructor call."""
        if self.python_code:
            return (eval_model, (self.python_code, self.params))
        return (self.__class__, (self._a, self._b, self._opstr))

    def get_components(self):
        if self._components is not None:
            return self._components
        if self._opstr in ('+', '*'):
            modifiers = []
            components = []
            first = self
            while isinstance(first, CombinedModel) and \
                first._opstr == self._opstr:
                second = first._b
                first = first._a
                if second.is_modifier():
                    modifiers.append(second)
                else:
                    components.append(second)
            if first.is_modifier():
                modifiers.append(first)
            else:
                components.append(first)
            ret = sum((c.get_components() for c in components), [])
            if modifiers:
                all_mods = reduce(lambda a, b: CombinedModel(a, b, self._opstr),
                                  modifiers)
                ret = [CombinedModel(all_mods, c, self._opstr) for c in ret]
        elif self._a.is_modifier():
            if self._b.is_modifier():
                # apparently nothing worthy of plotting
                ret = []
            else:
                ret = [CombinedModel(self._a, c, self._opstr)
                       for c in self._b.get_components()]
        elif self._b.is_modifier():
            ret = [CombinedModel(c, self._b, self._opstr)
                   for c in self._a.get_components()]
        else:
            # no modifiers
            ret = self._a.get_components() + self._b.get_components()
        self._components = ret
        return ret

    def get_description(self):
        if self.python_code:
            return self.python_code
        s = ''
        if isinstance(self._a, CombinedModel) and \
            self.op_prio[self._a._opstr] < self.op_prio[self._opstr]:
            s += '(%s)' % self._a.get_description()
        else:
            s += self._a.get_description()
        s += ' ' + self._opstr + ' '
        if isinstance(self._b, CombinedModel) and \
            self.op_prio[self._b._opstr] < self.op_prio[self._opstr]:
            s += '(%s)' % self._b.get_description()
        else:
            s += self._b.get_description()
        return s

    def get_pick_points(self):
        """Get a list of point names that should be picked for initial guess."""
        return self._a.get_pick_points() + self._b.get_pick_points()

    def convert_pick(self, *args):
        """Convert pick point coordinates (x,y) to parameter initial guesses."""
        npp = len(self._a.get_pick_points())
        d = self._a.convert_pick(*args[:npp])
        d.update(self._b.convert_pick(*args[npp:]))
        return d


class Constant(Model):
    """Constant function - no parameters.

    Used for math operations between models and numbers.  Not to be confused
    with the Const model from models.other.
    """
    def __init__(self, const):
        self.const = const
        self.fcn = lambda p, x: const

    def __reduce__(self):
        """Pickling support: reconstruct the object from a constructor call."""
        if self.python_code:
            return (eval_model, (self.python_code, self.params))
        return (self.__class__, (self.const,))

    def is_modifier(self):
        return True


class Function(Model):
    """Model using a function provided by the user.

    Parameters are extracted from the function's arguments and passed
    positionally.
    """
    def __init__(self, fcn, name=None, **init):
        self._real_fcn = fcn
        if name is None:
            if fcn.__name__ != '<lambda>':
                name = fcn.__name__
            else:
                name = ''
        pvs = self._init_params(name, inspect.getargspec(fcn)[0][1:], init)

        self.fcn = lambda p, x: \
            self._real_fcn(x, *(p[pv] for pv in pvs))

    def get_description(self):
        if self.python_code:
            return self.python_code
        return 'Function(%s, %s)' % (self.name, self._real_fcn.func_name)


class Custom(Model):
    """Create a model class from a user-defined expression."""
    def __init__(self, name, params, expr, **init):
        self._params = params
        self._expr = expr
        params = params.split()
        pvs = self._init_params(name, params, init)
        param_assign = ['%s = p[%r]' % pv for pv in zip(params, pvs)]
        namespace = param.expr_namespace.copy()
        exec '''def _fcn(p, x):
        %s
        return %s
        ''' % ('\n        '.join(param_assign), expr) in namespace
        self.fcn = namespace['_fcn']

    def get_description(self):
        return 'Custom(%r, %r, %r)' % (self.name, self._params, self._expr)


class GlobalModel(Model):
    """Model for a global fit for multiple datasets.

    Parameters can be global ("overall" parameters) or local to each dataset.
    Global parameters can be referenced in expressions from local parameters,
    but no the other way around.
    """

    def __init__(self, model, datas):
        self._model = model
        self._datas = datas
        ndata = len(datas)

        # generate a new parameter list with the model's original parameters
        # duplicated N times for N datasets, except for overall parameters;
        # the duplicates get named oldname__i where i is the data index

        self.params = []
        overall_params = []
        diff_params = [[] for i in range(ndata)]
        for p in model.params:
            if p.overall:
                self.params.append(p.copy())
                overall_params.append(p.name)
            else:
                for i in range(ndata):
                    new_param = p.copy(p.name + '__' + str(i))
                    self.params.append(new_param)
                    diff_params[i].append((p.name, new_param))

        # rewrite expressions to refer to the new parameter names (__i suffix)
        # and new data meta dictionaries (data.di)

        for i, dplist in enumerate(diff_params):
            for oldname, param in dplist:
                param._orig_expr = param.expr
                if not param.expr:
                    continue
                for oldname0, p0 in dplist:
                    param.expr = param.expr.replace(oldname0, p0.name)
                param.expr = data_re.sub('data.d%d' % i, param.expr)

        # global fitting function: call model function once for each dataset
        # with the original data, with the parameter values taken from the
        # duplicated params

        def new_fcn(p, x):
            results = []
            dpd = dict((pn, p[pn]) for pn in overall_params)
            for i, data in enumerate(datas):
                dpd.update((opn, p[pn.name]) for (opn, pn) in diff_params[i])
                results.append(model.fcn(dpd, data.fit_columns[0]))
            return concatenate(results)
        self.fcn = new_fcn

    def fit(self, datas, **kw):

        # fit a cumulative data set consisting of a concatenation of all data
        fitcols = [d.fit_columns for d in datas]

        cumulative_data = Dataset.from_arrays(
            'cumulative data',
            concatenate([cols[0] for cols in fitcols]),
            concatenate([cols[1] for cols in fitcols]),
            concatenate([cols[2] for cols in fitcols]),
            dict(('d%d' % i, d.meta) for (i, d) in enumerate(datas)),
        )
        overall_res = Model.fit(self, cumulative_data, **kw)

        # generate a list of results for each dataset with the original
        # parameter names and expressions

        reslist = []
        for i, data in enumerate(datas):
            suffix = '__%d' % i
            paramlist = []
            for p in self.params:
                if p.overall:
                    paramlist.append(p)
                elif p.name.endswith(suffix):
                    clone_param = p.copy(p.name[:-len(suffix)])
                    clone_param.expr = p._orig_expr
                    paramlist.append(clone_param)
            chi2 = get_chisqr(self._model.fcn, data.x, data.y, data.dy, paramlist)
            reslist.append(Result(overall_res.success, data, self._model,
                                  paramlist, overall_res.message, chi2))
        return MultiResult(reslist)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Models for different peak shapes."""

from numpy import exp, log, sqrt, sin, cos, pi
from scipy.special import wofz

from ufit.models import Model

__all__ = ['Gauss', 'GaussInt', 'Lorentz', 'LorentzInt',
           'Voigt', 'PseudoVoigt', 'DHO']


class Gauss(Model):
    """Gaussian peak

    Parameters:

    * `pos` - Peak center position
    * `ampl` - Amplitude at center
    * `fwhm` - Full width at half maximum
    """
    param_names = ['pos', 'ampl', 'fwhm']

    def __init__(self, name='', pos=None, ampl=None, fwhm=None):
        pp, pa, pf = self._init_params(name, self.param_names, locals())
        # amplitude and fwhm should be positive
        self.params[1].finalize = abs
        self.params[2].finalize = abs

        self.fcn = lambda p, x: \
            abs(p[pa]) * exp(-(x - p[pp])**2/p[pf]**2 * 4*log(2))

    pick_points = ['peak', 'width']

    def convert_pick(self, p, w):
        return {
            self.params[0].name: p[0],  # position
            self.params[1].name: p[1],  # peak amplitude
            self.params[2].name: 2*abs(w[0] - p[0]),  # FWHM
        }


class GaussInt(Model):
    """Gaussian peak with integrated intensity parameter

    Parameters:

    * `pos` - Peak center position
    * `int` - Integrated intensity
    * `fwhm` - Full width at half maximum
    """
    param_names = ['pos', 'int', 'fwhm']

    def __init__(self, name='', pos=None, int=None, fwhm=None):
        pp, pint, pf = self._init_params(name, self.param_names, locals())
        # integration and fwhm should be positive
        self.params[1].finalize = abs
        self.params[2].finalize = abs

        self.fcn = lambda p, x: \
            abs(p[pint]) / (abs(p[pf]) * sqrt(2*pi)) * exp(-(x - p[pp])**2/p[pf]**2 * 4*log(2))

    pick_points = ['peak', 'width']

    def convert_pick(self, p, w):
        fwhm = 2*abs(w[0] - p[0])
        return {
            self.params[0].name: p[0],  # position
            self.params[1].name: p[1] * fwhm * sqrt(2*pi),  # peak intensity (integrated)
            self.params[2].name: fwhm,  # FWHM
        }


class Lorentz(Model):
    """Lorentzian peak

    Parameters:

    * `pos` - Peak center position
    * `ampl` - Amplitude at center
    * `fwhm` - Full width at half maximum
    """
    param_names = ['pos', 'ampl', 'fwhm']

    def __init__(self, name='', pos=None, ampl=None, fwhm=None):
        pp, pa, pf = self._init_params(name, self.param_names, locals())
        # amplitude and fwhm should be positive
        self.params[1].finalize = abs
        self.params[2].finalize = abs

        self.fcn = lambda p, x: abs(p[pa]) / (1 + 4*(x - p[pp])**2/p[pf]**2)

    pick_points = ['peak', 'width']

    def convert_pick(self, p, w):
        return {
            self.params[0].name: p[0],  # position
            self.params[1].name: p[1],  # peak amplitude
            self.params[2].name: 2*abs(w[0] - p[0]),  # FWHM
        }


class LorentzInt(Model):
    """Lorentzian peak with integrated intensity parameter

    Parameters:

    * `pos` - Peak center position
    * `int` - Integrated intensity
    * `fwhm` - Full width at half maximum
    """
    param_names = ['pos', 'int', 'fwhm']

    def __init__(self, name='', pos=None, int=None, fwhm=None):
        pp, pint, pf = self._init_params(name, self.param_names, locals())
        # integration and fwhm should be positive
        self.params[1].finalize = abs
        self.params[2].finalize = abs

        self.fcn = lambda p, x: 2 * abs(p[pint]) / (pi * p[pf]) / (1 + 4*(x - p[pp])**2/p[pf]**2)

    pick_points = ['peak', 'width']

    def convert_pick(self, p, w):
        fwhm = 2*abs(w[0] - p[0])
        return {
            self.params[0].name: p[0],  # position
            self.params[1].name: p[1] * fwhm * pi/2,  # integrated intensity
            self.params[2].name: fwhm,  # FWHM
        }


class Voigt(Model):
    """Voigt peak

    A convolution of a Gaussian and a Lorentzian.

    Parameters:

    * `pos` - Peak center position
    * `ampl` - Amplitude at center
    * `fwhm` - Full width at half maximum of the Gauss part
    * `shape` - Lorentz contribution
    """
    param_names = ['pos', 'ampl', 'fwhm', 'shape']

    def __init__(self, name='', pos=None, ampl=None, fwhm=None, shape=None):
        pp, pa, pf, psh = self._init_params(name, self.param_names, locals())
        # amplitude and fwhms should be positive
        self.params[1].finalize = abs
        self.params[2].finalize = abs
        self.params[3].finalize = abs

        self.fcn = lambda p, x: \
            p[pa] / wofz(1j*sqrt(log(2))*p[psh]).real * \
            wofz(2*sqrt(log(2)) * (x-p[pp])/p[pf] + 1j*sqrt(log(2))*p[psh]).real

    pick_points = ['peak', 'width']

    def convert_pick(self, p, w):
        return {
            self.params[0].name: p[0],  # position
            self.params[1].name: p[1],  # peak amplitude
            self.params[2].name: 2*abs(w[0] - p[0]),  # FWHM of Gauss
            self.params[3].name: 0,
        }


class PseudoVoigt(Model):
    """Pseudo-Voigt peak

    A pseudo-convolution of a Gaussian and a Lorentzian.

    Parameters:

    * `pos` - Peak center position
    * `ampl` - Amplitude at center
    * `fwhm` - Full width at half maximum
    * `eta` - Lorentzicity
    """
    param_names = ['pos', 'ampl', 'fwhm', 'eta']

    def __init__(self, name='', pos=None, ampl=None, fwhm=None, eta=0.5):
        pp, pa, pf, pe = self._init_params(name, self.param_names, locals())
        # amplitude and fwhm should be positive
        self.params[1].finalize = abs
        self.params[2].finalize = abs
        # eta should be between 0 and 1
        self.params[3].finalize = lambda e: e % 1.0

        self.fcn = lambda p, x: abs(p[pa]) * \
            ((p[pe] % 1.0) / (1 + 4*(x - p[pp])**2/p[pf]**2) +
             (1-(p[pe] % 1.0)) * exp(-(x - p[pp])**2/p[pf]**2 * 4*log(2)))

    pick_points = ['peak', 'width']

    def convert_pick(self, p, w):
        return {
            self.params[0].name: p[0],  # position
            self.params[1].name: p[1],  # peak amplitude
            self.params[2].name: 2*abs(w[0] - p[0]),  # FWHM
        }


class DHO(Model):
    """Damped Harmonic Oscillator

    Two Lorentzians centered around zero with a common width and amplitude,
    respecting the Bose factor.

    Parameters:

    * `center` - Energy zero
    * `pos` - omega_0
    * `ampl` - Amplitude
    * `gamma` - Damping
    * `tt` - Temperature in K
    """
    param_names = ['center', 'pos', 'ampl', 'gamma', 'tt']

    def __init__(self, name='',
                 center=0, pos=None, ampl=None, gamma=None, tt=None):
        pc, pp, pa, pg, ptt = self._init_params(name, self.param_names,
                                                locals())
        # pos, amplitude and gamma should be positive
        self.params[1].finalize = abs
        self.params[2].finalize = abs
        self.params[3].finalize = abs
        self.fcn = lambda p, x: x / (1. - exp(-11.6045*(x+0.00001) / p[ptt])) * \
            abs(p[pa]) * abs(p[pg]) / \
            ((p[pp]**2 - (x - p[pc])**2)**2 + (p[pg]*(x - p[pc]))**2)

    pick_points = ['left peak', 'width of left peak', 'right peak']

    def convert_pick(self, p1, w, p2):
        return {
            self.params[0].name: 0.5*(p1[0] + p2[0]),  # center
            self.params[1].name: 0.5*abs(p1[0] - p2[0]),  # position
            self.params[2].name: p1[1] * 0.01,  # peak amplitude
            self.params[3].name: 2*abs(w[0] - p1[0]),  # gamma
        }


class Gauss2D(Model):
    """Gaussian peak in two dimensions

    Parameters:

    * `bkgd`   - Background
    * `pos_x`  - X center position
    * `pos_y`  - Y center position
    * `ampl`   - amplitude
    * `fwhm_x` - Full width in X direction
    * `fwhm_y` - Full width in Y direction
    * `theta`  - rotation of Gaussian in radians
    """
    param_names = ['bkgd', 'pos_x', 'pos_y', 'ampl', 'fwhm_x', 'fwhm_y', 'theta']

    def __init__(self, name='', bkgd=None, pos_x=None, pos_y=None, ampl=None,
                 fwhm_x=None, fwhm_y=None, theta=None):
        pb, ppx, ppy, pa, pfx, pfy, pth = self._init_params(
            name, self.param_names, locals())
        self.params[3].finalize = abs
        self.params[4].finalize = abs
        self.params[5].finalize = abs

        def fcn(p, x):
            # rotate coordinate system by theta
            c, s = cos(p[pth]), sin(p[pth])
            x1 = (x[:,0] - p[ppx])*c - (x[:,1] - p[ppy])*s
            y1 = (x[:,0] - p[ppx])*s + (x[:,1] - p[ppy])*c
            return abs(p[pb]) + abs(p[pa]) * \
                exp(-x1**2/p[pfx]**2 * 4*log(2)) * \
                exp(-y1**2/p[pfy]**2 * 4*log(2))
        self.fcn = fcn




#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data fitter panel."""

from PyQt4.QtCore import SIGNAL, Qt
from PyQt4.QtGui import QApplication, QWidget, QMainWindow, QGridLayout, \
    QFrame, QLabel, QDialogButtonBox, QCheckBox, QMessageBox, QSplitter, \
    QComboBox, QKeySequence, QIcon

from ufit.gui import logger
from ufit.gui.common import loadUi, MPLCanvas, MPLToolbar, SmallLineEdit
from ufit.gui.session import session


def is_float(x):
    try:
        float(x)
        return True
    except ValueError:
        return False


class Fitter(QWidget):

    def __init__(self, parent, standalone=False, fit_kws={}):
        QWidget.__init__(self, parent)
        self.item = getattr(parent, 'item', None)
        self.logger = logger.getChild('fitter')
        self.picking = None
        self.last_result = None
        self.model = None
        self.data = None
        self.param_controls = {}
        self.fit_kws = fit_kws

        self.standalone = standalone
        self.createUI(standalone)

    def createUI(self, standalone):
        loadUi(self, 'fitter.ui')

        if standalone:
            self.buttonBox.addButton(QDialogButtonBox.Close)
        self.buttonBox.addButton('Initial guess', QDialogButtonBox.HelpRole)
        self.buttonBox.addButton('Save params', QDialogButtonBox.ResetRole)
        self.buttonBox.addButton('Restore saved', QDialogButtonBox.ResetRole)
        self.buttonBox.addButton('Replot', QDialogButtonBox.ActionRole)
        fitbtn = self.buttonBox.addButton('Fit', QDialogButtonBox.ApplyRole)
        fitbtn.setShortcut(QKeySequence('Ctrl+F'))
        fitbtn.setIcon(QIcon.fromTheme('dialog-ok'))

    def initialize(self, model, data, fit=True, keep_old=True):
        self.picking = None
        self.last_result = None

        old_model = self.model
        self.modelLabel.setText(model.get_description())
        self.data = data
        self.model = model

        self.param_controls = {}
        self.create_param_controls()

        # try to transfer values of old parameters to new
        if keep_old and old_model is not None:
            oldp_dict = dict((p.name, p) for p in old_model.params)
            self.restore_from_params(oldp_dict)

        self.connect(session, SIGNAL('modelFitted'), self.on_modelFitted)

        if self.standalone:
            if fit:
                self.do_fit()
            else:
                self.emit(SIGNAL('replotRequest'), None)

    def create_param_controls(self):
        self.param_controls = {}
        self.param_frame = QFrame(self)
        layout = QGridLayout()
        for j, text in enumerate(('Param', 'Value', 'Error', 'Fix', 'Expr',
                                  'Min', 'Max', 'Delta')):
            ctl = QLabel(text, self)
            ctl.setFont(self.statusLabel.font())
            layout.addWidget(ctl, 0, j)
        i = 1
        self.original_params = {}
        combo_items = [''] + [par.name for par in self.model.params] + \
            ['data.' + m for m in sorted(self.data.meta)
             if isinstance(self.data.meta[m], (int, long, float))]
        for p in self.model.params:
            e0 = QLabel(p.name, self)
            e1 = SmallLineEdit('%.5g' % p.value, self)
            e2 = QLabel(u'± %.5g' % p.error, self)
            e3 = QCheckBox(self)
            e4 = QComboBox(self)
            e4.setEditable(True)
            e4.addItems(combo_items)
            if p.expr and is_float(p.expr):
                e1.setText(p.expr)
                e3.setChecked(True)
                e4.lineEdit().setText('')
            else:
                e4.lineEdit().setText(p.expr or '')
            e5 = SmallLineEdit(p.pmin is not None and '%.5g' % p.pmin or '', self)
            e6 = SmallLineEdit(p.pmax is not None and '%.5g' % p.pmax or '', self)
            e7 = SmallLineEdit(p.delta and '%.5g' % p.delta or '', self)
            ctls = self.param_controls[p] = (e0, e1, e2, e3, e4, e5, e6, e7)
            for j, ctl in enumerate(ctls):
                layout.addWidget(ctl, i, j)
            i += 1
            self.original_params[p.name] = p.copy()
            self.connect(e1, SIGNAL('returnPressed()'), self.do_plot)
            self.connect(e4.lineEdit(), SIGNAL('returnPressed()'), self.do_plot)
            self.connect(e5, SIGNAL('returnPressed()'), self.do_plot)
            self.connect(e6, SIGNAL('returnPressed()'), self.do_plot)
            self.connect(e3, SIGNAL('clicked(bool)'), self.update_enables)
            self.connect(e4, SIGNAL('editTextChanged(const QString&)'),
                         self.update_enables)
        layout.setRowStretch(i+1, 1)
        self.param_frame.setLayout(layout)
        self.param_scroll.setWidget(self.param_frame)
        self.update_enables()

    def update_enables(self, *ignored):
        for p, ctls in self.param_controls.iteritems():
            # if there is an expr...
            if ctls[4].currentText():
                # disable value and minmax, check "fixed" and disable "fixed"
                ctls[1].setEnabled(False)
                ctls[3].setCheckState(Qt.PartiallyChecked)  # implicitly fixed
                ctls[3].setEnabled(False)
                ctls[5].setEnabled(False)
                ctls[6].setEnabled(False)
                ctls[7].setEnabled(False)
            # else, if "fixed" is checked...
            elif ctls[3].checkState() == Qt.Checked:
                # enable value, but disable expr and minmax
                ctls[1].setEnabled(True)
                ctls[4].setEnabled(False)
                ctls[5].setEnabled(False)
                ctls[6].setEnabled(False)
                ctls[7].setEnabled(False)
            # else: not fixed, no expr
            else:
                # enable everything
                ctls[1].setEnabled(True)
                ctls[3].setEnabled(True)
                ctls[3].setCheckState(Qt.Unchecked)
                ctls[3].setTristate(False)
                ctls[4].setEnabled(True)
                ctls[5].setEnabled(True)
                ctls[6].setEnabled(True)
                ctls[7].setEnabled(True)

    def on_buttonBox_clicked(self, button):
        role = self.buttonBox.buttonRole(button)
        if role == QDialogButtonBox.RejectRole:
            self.emit(SIGNAL('closeRequest'))
        elif role == QDialogButtonBox.ApplyRole:
            self.do_fit()
        elif role == QDialogButtonBox.ActionRole:
            self.do_plot()
        elif role == QDialogButtonBox.HelpRole:
            self.do_pick()
        else:
            if button.text() == 'Save params':
                self.save_original_params()
                self.statusLabel.setText('Current parameter values saved.')
            else:
                self.restore_from_params(self.original_params)
                self.statusLabel.setText('Saved parameter values restored.')

    def update_from_controls(self):
        for p, ctls in self.param_controls.iteritems():
            _, val, _, fx, expr, pmin, pmax, delta = ctls
            p.value = float(val.text()) if val.text() else 0
            if fx.checkState() == Qt.Checked:
                p.expr = str(val.text())
            else:
                p.expr = str(expr.currentText())
            p.pmin = float(pmin.text()) if pmin.text() else None
            p.pmax = float(pmax.text()) if pmax.text() else None
            p.delta = float(delta.text()) if delta.text() else 0
        self.update_enables()
        session.set_dirty()

    def restore_from_params(self, other_params):
        for p in self.model.params:
            if p.name not in other_params:
                continue
            p0 = other_params[p.name]
            ctls = self.param_controls[p]
            ctls[1].setText('%.5g' % p0.value)
            ctls[2].setText(u'± %.5g' % p0.error)
            ctls[3].setChecked(False)
            if p0.expr and is_float(p0.expr):
                ctls[1].setText(p0.expr)
                ctls[3].setChecked(True)
                ctls[4].lineEdit().setText('')
            else:
                ctls[4].lineEdit().setText(p0.expr or '')
            ctls[5].setText(p0.pmin is not None and '%.5g' % p0.pmin or '')
            ctls[6].setText(p0.pmax is not None and '%.5g' % p0.pmax or '')
            ctls[7].setText(p0.delta and '%.5g' % p0.delta or '')
        session.set_dirty()
        self.do_plot()

    def save_original_params(self):
        self.original_params = {}
        for p in self.model.params:
            self.original_params[p.name] = p.copy()

    def on_canvas_pick(self, event):
        if not self.picking:
            return
        if not hasattr(event, 'xdata') or event.xdata is None:
            return
        self._pick_values.append((event.xdata, event.ydata))
        if len(self._pick_values) == len(self._pick_points):
            self.picking = False
            self.statusLabel.setText('')
            self._pick_finished()
        else:
            self.statusLabel.setText('%s: click on %s' %
                (self.picking, self._pick_points[len(self._pick_values)]))

    def do_pick(self, *args):
        if self.picking:
            return
        self.emit(SIGNAL('pickRequest'), self)
        self._pick_points = self.model.get_pick_points()
        self._pick_values = []
        self.picking = 'Guess'
        self.statusLabel.setText('Guess: click on %s' % self._pick_points[0])
        def callback():
            self.model.apply_pick(self._pick_values)
            for p in self.model.params:
                ctls = self.param_controls[p]
                if not p.expr:
                    ctls[1].setText('%.5g' % p.value)
            session.set_dirty()
            self.do_plot()
        self._pick_finished = callback

    def do_plot(self, *ignored):
        self.update_from_controls()
        self.emit(SIGNAL('replotRequest'), None)

    def do_fit(self):
        if self.picking:
            QMessageBox.information(self, 'Fitting',
                                    'Please finish the picking operation first.')
            return
        self.update_from_controls()
        self.statusLabel.setText('Working...')
        self.statusLabel.repaint()
        QApplication.processEvents()
        try:
            res = self.model.fit(self.data, **self.fit_kws)
        except Exception, e:
            self.logger.exception('Error during fit')
            self.statusLabel.setText('Error during fit: %s' % e)
            return
        self.on_modelFitted(self.item, res)

        self.emit(SIGNAL('replotRequest'), True)
        session.set_dirty()

    def on_modelFitted(self, item, res):
        if item is not self.item:
            return

        self.statusLabel.setText((res.success and 'Converged. ' or 'Failed. ')
                                 + res.message +
                                 ' Reduced chi^2 = %.3g.' % res.chisqr)

        for p in res.params:
            self.param_controls[p][1].setText('%.5g' % p.value)
            self.param_controls[p][2].setText(u'± %.5g' % p.error)

        self.last_result = res


class FitterMain(QMainWindow):
    def __init__(self, model, data, fit=True, fit_kws={}):
        QMainWindow.__init__(self)
        layout = QSplitter(Qt.Vertical, self)
        self.canvas = MPLCanvas(self)
        self.toolbar = MPLToolbar(self.canvas, self)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        self.fitter = Fitter(self, standalone=True, fit_kws=fit_kws)
        self.canvas.mpl_connect('button_press_event',
                                self.fitter.on_canvas_pick)
        self.connect(self.fitter, SIGNAL('closeRequest'), self.close)
        self.connect(self.fitter, SIGNAL('replotRequest'), self.replot)
        self.fitter.initialize(model, data, fit)
        layout.addWidget(self.fitter)
        self.setCentralWidget(layout)
        self.setWindowTitle('Fitting: data %s' % data.name)

    def replot(self, limits=True):
        plotter = self.canvas.plotter
        plotter.reset(limits=limits)
        try:
            plotter.plot_data(self.fitter.data)
            plotter.plot_model_full(self.fitter.model, self.fitter.data)
        except Exception:
            logger.exception('Error while plotting')
        else:
            plotter.draw()


def start(model, data, fit=True, **fit_kws):
    app = QApplication([])
    win = FitterMain(model, data, fit, fit_kws)
    win.show()
    app.exec_()
    return win.fitter.last_result



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data loader object."""

from numpy import ones, sqrt

from ufit import UFitError
from ufit.data.dataset import ScanData, ImageData, DataList, DatasetList


class Loader(object):
    def __init__(self):
        self.format = 'auto'
        self.template = '%d'
        self.sets = DataList()

    def _get_reader(self, filename, fobj):
        from ufit.data import data_formats, data_formats_image
        if self.format == 'auto':
            for n, m in data_formats.iteritems():
                # check 'simple' last
                if n != 'simple' and m.check_data(fobj):
                    return m, n in data_formats_image
            if data_formats['simple'].check_data(fobj):
                return data_formats['simple'], False
            raise UFitError('File %r has no recognized file format' % filename)
        return data_formats[self.format], self.format in data_formats_image

    def _inner_load(self, n, xcol, ycol, dycol=None, ncol=None, nscale=1):
        try:
            filename = self.template % n
        except TypeError:
            filename = self.template
        fobj = open(filename, 'rb')
        rdr, isimg = self._get_reader(filename, fobj)
        if isimg:
            return self._inner_load_image(rdr, filename, fobj, n, ncol, nscale)
        return self._inner_load_scan(rdr, filename, fobj, n, xcol, ycol, dycol,
                                     ncol, nscale)

    def _inner_load_image(self, rdr, filename, fobj, n, ncol, nscale):
        arr, darr, meta = rdr.read_data(filename, fobj)
        if 'filenumber' not in meta:
            meta['filenumber'] = n
        meta['datafilename'] = filename
        if ncol == 'auto':
            ncol = rdr.guess_norm(meta)
        if ncol is not None:
            norm = meta[ncol]
        else:
            norm = 1
        dset = ImageData(meta, arr, darr, norm, nscale)
        self.sets[n] = dset
        return dset

    def _inner_load_scan(self, rdr, filename, fobj, n,
                         xcol, ycol, dycol, ncol, nscale):
        colnames, coldata, meta = rdr.read_data(filename, fobj)
        colguess = rdr.guess_cols(colnames, coldata, meta)
        if 'filenumber' not in meta:
            meta['filenumber'] = n
        meta['datafilename'] = filename
        for colname, colvalues in zip(colnames, coldata.T):
            meta['col_%s' % colname] = colvalues
        datarr = ones((len(coldata), 4))

        def colindex(col):
            if isinstance(col, str):
                try:
                    return colnames.index(col)
                except ValueError:
                    raise UFitError('No such data column: %s' % col)
            elif 1 <= col <= len(colnames):
                return col - 1   # 1-based indices
            else:
                raise UFitError('Data has only %d columns (but column %s is '
                                'requested)' % (len(colnames), col))
        use_hkl = False
        if xcol == 'hkl':
            xcol = 'auto'
            use_hkl = True
        if xcol == 'auto':
            xcol = colguess[0]
        datarr[:,0] = coldata[:,colindex(xcol)]
        if ycol == 'auto':
            ycol = colguess[1]
        datarr[:,1] = coldata[:,colindex(ycol)]
        if dycol == 'auto':
            dycol = colguess[2]
        if dycol is not None:
            datarr[:,2] = coldata[:,colindex(dycol)]
        else:
            datarr[:,2] = sqrt(datarr[:,1])
        if ncol == 'auto':
            ncol = colguess[3]
        if ncol is not None:
            datarr[:,3] = coldata[:,colindex(ncol)]
            if nscale == -1:
                nscale = int(float('%.2g' % datarr[:,3].mean()))

        def colname(col):
            if col is None:
                return None
            elif isinstance(col, str):
                return col
            return colnames[col - 1]   # 1-based indices
        if use_hkl:
            meta['is_hkldata'] = True
        dset = ScanData(meta, datarr, colname(xcol), colname(ycol),
                        colname(ncol), nscale)
        if use_hkl and 'hkle' in dset.meta:  # 3-axis support
            dset.x = dset.meta['hkle']
        self.sets[n] = dset
        return dset

    def load(self, n, xcol, ycol, dycol=None, ncol=None, nscale=1):
        try:
            return self._inner_load(n, xcol, ycol, dycol, ncol, nscale)
        except Exception, e:
            raise UFitError('Could not load data file %d: %s' % (n, e))

    def guess_cols(self, n):
        try:
            filename = self.template % n
        except TypeError:
            filename = self.template
        fobj = open(filename, 'rb')
        rdr, isimg = self._get_reader(filename, fobj)
        if isimg:
            arr, darr, meta = rdr.read_data(filename, fobj)
            mguess = rdr.guess_norm(meta)
            if mguess:
                nmon = int(float('%.2g' % meta[mguess]))
            else:
                nmon = 1
            xguess, yguess, dyguess = None, None, None
            colnames = meta.keys()
        else:
            colnames, coldata, meta = rdr.read_data(filename, fobj)
            xguess, yguess, dyguess, mguess = rdr.guess_cols(colnames, coldata, meta)
            if mguess is not None:
                # use average monitor counts for normalization, but
                # round to 2 significant digits
                moncol = coldata[:,colnames.index(mguess)]
                nmon = int(float('%.2g' % moncol.mean()))
            else:
                nmon = 0
            if yguess is None and len(colnames) > 1:
                yguess = colnames[1]
                if len(colnames) > 2:
                    dyguess = colnames[2]
        return colnames, xguess, yguess, dyguess, mguess, nmon

    def load_numors(self, nstring, binsize, xcol, ycol, dycol=None,
                    ncol=None, nscale=1, floatmerge=False):
        """Load a number of data files and merge them according to numor
        list operations:

        * ``,`` - put single files in individual data sets
        * ``-`` - put sequential files in individual data sets
        * ``+`` - merge single files
        * ``>`` - merge sequential files
        """
        if not isinstance(binsize, (int, float)):
            raise UFitError('binsize argument must be a number')
        def toint(a):
            try:
                return int(a)
            except ValueError:
                raise UFitError('Invalid file number: %r' % a)
        # operator "precedence": ',' has lowest, then '+',
        # then '-' and '>' (equal)
        parts1 = nstring.split(',')
        datasets = []
        for part1 in parts1:
            if '-' in part1:
                a, b = map(toint, part1.split('-'))
                datasets.extend(
                    self.load(n, xcol, ycol, dycol, ncol, nscale).merge(
                        binsize, floatmerge=floatmerge) for n in range(a, b+1))
            else:
                parts2 = part1.split('+')
                inner = []
                for part2 in parts2:
                    if '>' in part2:
                        a, b = map(toint, part2.split('>'))
                        ds = [self.load(n, xcol, ycol, dycol, ncol, nscale)
                              for n in range(a, b+1)]
                        inner.append(ds[0].merge(binsize, *ds[1:],
                                                 floatmerge=floatmerge))
                    else:
                        inner.append(
                            self.load(toint(part2), xcol, ycol, dycol,
                                      ncol, nscale))
                datasets.append(inner[0].merge(binsize, *inner[1:],
                                               floatmerge=floatmerge))
        return DatasetList(datasets)
